import logging
import re

from behave import then, when

from features.steps.files import when_i_create_file_with_content
from features.steps.packages import when_i_apt_install, when_i_apt_update
from features.steps.shell import when_i_run_command, when_i_run_shell_command
from features.util import (
    ALL_BINARY_PACKAGE_NAMES,
    NORMAL_BINARY_PACKAGE_NAMES,
    SUT,
    InstallationSource,
    build_debs,
    get_debs_for_series,
)

SETUP_PRO_PACKAGE_SOURCES_SCRIPTS = {
    InstallationSource.ARCHIVE: "sudo apt update",
    InstallationSource.DAILY: """\
sudo add-apt-repository ppa:ua-client/daily
sudo apt update""",
    InstallationSource.STAGING: """\
sudo add-apt-repository ppa:ua-client/staging
sudo apt update""",
    InstallationSource.STABLE: """\
sudo add-apt-repository ppa:ua-client/stable
sudo apt update""",
    InstallationSource.PROPOSED: """\
cat > /etc/apt/sources.list.d/proposed.list << EOF
deb http://archive.ubuntu.com/ubuntu/ {series}-proposed main
EOF
cat > /etc/apt/preferences.d/lower-proposed << EOF
Package: *
Pin: release a=*-proposed
Pin-Priority: 400
EOF
cat > /etc/apt/preferences.d/upper-pro-posed << EOF
Package: {packages}
Pin: release a=*-proposed
Pin-Priority: 1001
EOF
sudo apt update""",
    InstallationSource.CUSTOM: """\
sudo add-apt-repository {ppa}
sudo apt update""",
}


def get_setup_pro_package_sources_script(context, series):
    script = SETUP_PRO_PACKAGE_SOURCES_SCRIPTS.get(
        context.pro_config.install_from, ""
    )
    script = script.format(
        ppa=context.pro_config.custom_ppa,
        series=series,
        packages=" ".join(ALL_BINARY_PACKAGE_NAMES),
    )
    return script


def setup_pro_package_sources(context, machine_name=SUT):
    instance = context.machines[machine_name].instance
    series = context.machines[machine_name].series
    script = get_setup_pro_package_sources_script(context, series)
    context.text = script
    when_i_create_file_with_content(
        context,
        "/tmp/setup_pro.sh",
        machine_name=machine_name,
    )
    instance.execute("sudo bash /tmp/setup_pro.sh")


@when("I install ubuntu-advantage-tools")
def when_i_install_uat(context, machine_name=SUT):
    instance = context.machines[machine_name].instance
    series = context.machines[machine_name].series
    is_pro = "pro" in context.machines[machine_name].machine_type
    setup_pro_package_sources(context, machine_name)

    if context.pro_config.install_from is InstallationSource.PREBUILT:
        debs = get_debs_for_series(context.pro_config.debs_path, series)
        logging.info("using debs: {}".format(debs))
        to_install = []
        for deb_name, deb_path in debs.non_cloud_pro_image_debs(series):
            instance_tmp_path = "/tmp/behave_{}.deb".format(deb_name)
            instance.push_file(deb_path, instance_tmp_path)
            to_install.append(instance_tmp_path)
        if is_pro:
            for deb_name, deb_path in debs.cloud_pro_image_debs():
                instance_tmp_path = "/tmp/behave_{}.deb".format(deb_name)
                instance.push_file(deb_path, instance_tmp_path)
                to_install.append(instance_tmp_path)
        when_i_apt_install(
            context, " ".join(to_install), machine_name=machine_name
        )
    elif context.pro_config.install_from is InstallationSource.LOCAL:
        debs = build_debs(
            series,
            sbuild_output_to_terminal=context.pro_config.sbuild_output_to_terminal,  # noqa: E501
        )
        to_install = []
        for deb_name, deb_path in debs.non_cloud_pro_image_debs(series):
            instance_tmp_path = "/tmp/behave_{}.deb".format(deb_name)
            instance.push_file(deb_path, instance_tmp_path)
            to_install.append(instance_tmp_path)
        if is_pro:
            for deb_name, deb_path in debs.cloud_pro_image_debs():
                instance_tmp_path = "/tmp/behave_{}.deb".format(deb_name)
                instance.push_file(deb_path, instance_tmp_path)
                to_install.append(instance_tmp_path)
        when_i_apt_install(
            context, " ".join(to_install), machine_name=machine_name
        )
    else:
        if series in ("xenial", "bionic", "focal", "jammy"):
            when_i_apt_install(
                context,
                "ubuntu-pro-client ubuntu-advantage-tools",
                machine_name=machine_name,
            )
        else:
            when_i_apt_install(
                context,
                "ubuntu-pro-client",
                machine_name=machine_name,
            )

        if is_pro:
            when_i_apt_install(
                context,
                "ubuntu-pro-auto-attach",
                machine_name=machine_name,
            )


@when("I install ubuntu-advantage-tools on the `{guest_name}` lxd guest")
def when_i_install_uat_on_lxd_guest(context, guest_name):
    # This function assumes "when_i_install_uat" was run on the SUT
    if context.pro_config.install_from in {
        InstallationSource.PREBUILT,
        InstallationSource.LOCAL,
    }:
        to_install = []
        for deb_name in NORMAL_BINARY_PACKAGE_NAMES:
            deb_file_name = "behave_{}.deb".format(deb_name)
            instance_tmp_path = "/tmp/{}".format(deb_file_name)
            guest_path = "/root/{}".format(deb_file_name)
            when_i_run_command(
                context,
                "lxc file push {tmp_path} {guest_name}{guest_path}".format(
                    tmp_path=instance_tmp_path,
                    guest_name=guest_name,
                    guest_path=guest_path,
                ),
                "with sudo",
            )
            to_install.append(guest_path)
        when_i_run_command(
            context,
            "lxc exec {guest_name} -- apt install -y {packages}".format(
                guest_name=guest_name, packages=" ".join(to_install)
            ),
            "with sudo",
        )
    else:
        setup_pro_package_sources(context)

        when_i_run_command(
            context,
            "lxc file push /tmp/setup_pro.sh {guest_name}/root/setup_pro.sh".format(  # noqa: E501
                guest_name=guest_name
            ),
            "with sudo",
        )
        when_i_run_command(
            context,
            "lxc exec {guest_name} -- bash /root/setup_pro.sh".format(
                guest_name=guest_name
            ),
            "with sudo",
        )
        when_i_run_command(
            context,
            "lxc exec {guest_name} -- apt install -y {packages}".format(
                guest_name=guest_name,
                packages=" ".join(NORMAL_BINARY_PACKAGE_NAMES),
            ),
            "with sudo",
        )


@when("I ensure -proposed is not enabled anymore")
def when_i_ensure_proposed_not_enabled(context, machine_name=SUT):
    if context.pro_config.install_from is InstallationSource.PROPOSED:
        when_i_run_command(
            context,
            "rm -f /etc/apt/sources.list.d/proposed.list",
            "with sudo",
            machine_name=machine_name,
        )
        when_i_apt_update(context, machine_name=machine_name)


@when("I have the `{series}` debs under test in `{dest}`")
def when_i_have_the_debs_under_test(context, series, dest):
    if context.pro_config.install_from is InstallationSource.LOCAL:
        debs = build_debs(
            series,
            sbuild_output_to_terminal=context.pro_config.sbuild_output_to_terminal,  # noqa: E501
        )

        for deb_name, deb_path in debs.all_debs(series):
            context.machines[SUT].instance.push_file(
                deb_path, "{}/{}.deb".format(dest, deb_name)
            )
    else:
        if context.pro_config.install_from is InstallationSource.PROPOSED:
            ppa_opts = ""
        else:
            if context.pro_config.install_from is InstallationSource.DAILY:
                ppa = "ppa:ua-client/daily"
            elif context.pro_config.install_from is InstallationSource.STAGING:
                ppa = "ppa:ua-client/staging"
            elif context.pro_config.install_from is InstallationSource.STABLE:
                ppa = "ppa:ua-client/stable"
            elif context.pro_config.install_from is InstallationSource.CUSTOM:
                ppa = context.pro_config.custom_ppa
                if not ppa.startswith("ppa"):
                    # assumes format "http://domain.name/user/ppa/ubuntu"
                    match = re.match(r"https?://[\w.]+/([^/]+/[^/]+)", ppa)
                    if not match:
                        raise AssertionError(
                            "ppa is in unsupported format: {}".format(ppa)
                        )
                    ppa = "ppa:{}".format(match.group(1))
            ppa_opts = "--distro ppa --ppa {}".format(ppa)
        download_cmd = "pull-lp-debs {} ubuntu-advantage-tools {}".format(
            ppa_opts, series
        )
        when_i_run_command(
            context, "apt-get install -y ubuntu-dev-tools", "with sudo"
        )
        when_i_run_command(context, download_cmd, "with sudo")
        logging.info("Download command `{}`".format(download_cmd))
        logging.info("stdout: {}".format(context.process.stdout))
        logging.info("stderr: {}".format(context.process.stderr))
        for package in ALL_BINARY_PACKAGE_NAMES:
            when_i_run_shell_command(
                context,
                "cp {package}_*.deb {package}.deb".format(package=package),
                "with sudo",
            )


@when(
    "I prepare the local PPAs to upgrade from `{release}` to `{next_release}`"
)
def when_i_create_local_ppas(context, release, next_release):
    if context.pro_config.install_from is not InstallationSource.LOCAL:
        return

    from features.steps.machines import given_a_machine

    # We need Kinetic or greater to support zstd when creating the PPAs
    given_a_machine(context, "noble", "lxd-container", machine_name="ppa")
    when_i_run_command(
        context, "apt-get update", "with sudo", machine_name="ppa"
    )
    when_i_run_command(
        context, "apt-get install -y aptly", "with sudo", machine_name="ppa"
    )
    create_local_ppa(context, release)
    create_local_ppa(context, next_release)
    repo_line = "deb [trusted=yes] http://{}:8080 {} main".format(
        context.machines["ppa"].instance.ip, release
    )
    repo_file = "/etc/apt/sources.list.d/local-ua.list"
    when_i_run_shell_command(
        context, "printf '{}\n' > {}".format(repo_line, repo_file), "with sudo"
    )
    when_i_run_command(
        context,
        "sh -c 'nohup aptly serve > /dev/null 2>&1 &'",
        "with sudo",
        machine_name="ppa",
    )


def create_local_ppa(context, release):
    when_i_run_command(
        context,
        "aptly repo create -distribution {} repo-{}".format(release, release),
        "with sudo",
        machine_name="ppa",
    )
    debs = build_debs(
        release,
        sbuild_output_to_terminal=context.pro_config.sbuild_output_to_terminal,
    )
    for deb_name, deb_path in debs.all_debs(release):
        deb_destination = "/tmp/{}.deb".format(deb_name)
        context.machines["ppa"].instance.push_file(deb_path, deb_destination)
        when_i_run_command(
            context,
            "aptly repo add repo-{} {}".format(release, deb_destination),
            "with sudo",
            machine_name="ppa",
        )
    when_i_run_command(
        context,
        "aptly publish repo -skip-signing repo-{}".format(release),
        "with sudo",
        machine_name="ppa",
    )


@when("I install ubuntu-pro-auto-attach")
@when("I install ubuntu-advantage-pro")
def when_i_install_pro(context, machine_name=SUT):
    if context.pro_config.install_from is InstallationSource.LOCAL:
        series = context.machines[machine_name].series
        debs = build_debs(
            series,
            sbuild_output_to_terminal=context.pro_config.sbuild_output_to_terminal,  # noqa: E501
        )

        to_install = []
        for deb_name, deb_path in debs.cloud_pro_image_debs():
            instance_tmp_path = "/tmp/behave_{}.deb".format(deb_name)
            context.machines[machine_name].instance.push_file(
                deb_path, instance_tmp_path
            )
            to_install.append(instance_tmp_path)
        when_i_apt_install(
            context, " ".join(to_install), machine_name=machine_name
        )
    else:
        when_i_apt_install(
            context, "ubuntu-pro-auto-attach", machine_name=machine_name
        )


APT_POLICY_IS = "the apt-cache policy of ubuntu-pro-client is"


@then(APT_POLICY_IS)
def then_i_apt_cache_policy_is(context):
    pass


@when("I check the apt-cache policy of ubuntu-pro-client")
def when_i_check_apt_cache_policy(context):
    when_i_run_command(context, "apt-get update", "with sudo")
    when_i_run_command(
        context, "apt-cache policy ubuntu-pro-client", "with sudo"
    )
    for step in context.scenario.steps:
        if step.name == APT_POLICY_IS:
            step.text = context.process.stdout


@when("I install transition package ubuntu-advantage-tools")
def when_i_install_transition_uat(context, machine_name=SUT):
    is_pro = "pro" in context.machines[machine_name].machine_type
    setup_pro_package_sources(context, machine_name)

    when_i_apt_install(
        context, "ubuntu-advantage-tools", machine_name=machine_name
    )
    if is_pro:
        when_i_apt_install(
            context,
            "ubuntu-advantage-pro",
            machine_name=machine_name,
        )



from collections import defaultdict
from datetime import datetime, timezone
from enum import Enum
from functools import lru_cache
from random import choice
from typing import Any, DefaultDict, Dict, List, Tuple

import apt_pkg  # type: ignore

from uaclient import exceptions, livepatch, messages, util
from uaclient.api.u.pro.security.status.reboot_required.v1 import (
    _reboot_required,
)
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.apt import (
    PreserveAptCfg,
    get_apt_cache_datetime,
    get_apt_pkg_cache,
    get_esm_apt_pkg_cache,
    get_pkg_candidate_version,
)
from uaclient.config import UAConfig
from uaclient.entitlements import ESMAppsEntitlement, ESMInfraEntitlement
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    ApplicationStatus,
    ContractStatus,
)
from uaclient.system import (
    get_distro_info,
    get_kernel_info,
    get_release_info,
    is_current_series_lts,
    is_supported,
)

ESM_SERVICES = ("esm-infra", "esm-apps")


class UpdateStatus(Enum):
    "Represents the availability of a security package."
    AVAILABLE = "upgrade_available"
    AVAILABLE_NOT_PREFERRED = "upgrade_available_not_preferred"
    UNATTACHED = "pending_attach"
    NOT_ENABLED = "pending_enable"
    UNAVAILABLE = "upgrade_unavailable"


def print_package_list(packages):
    print(util.create_package_list_str(packages))


@lru_cache(maxsize=None)
def get_origin_information_to_service_map():
    series = get_release_info().series
    return {
        ("Ubuntu", "{}-security".format(series)): "standard-security",
        ("UbuntuESMApps", "{}-apps-security".format(series)): "esm-apps",
        ("UbuntuESM", "{}-infra-security".format(series)): "esm-infra",
        ("UbuntuESMApps", "{}-apps-updates".format(series)): "esm-apps",
        ("UbuntuESM", "{}-infra-updates".format(series)): "esm-infra",
    }


def get_installed_packages_by_origin() -> (
    DefaultDict["str", List[apt_pkg.Package]]
):
    result = defaultdict(list)

    with PreserveAptCfg(get_apt_pkg_cache) as cache:
        installed_packages = [
            package for package in cache.packages if package.current_ver
        ]
        result["all"] = installed_packages

        dep_cache = apt_pkg.DepCache(cache)

        for package in installed_packages:
            result[
                get_origin_for_installed_package(package, dep_cache)
            ].append(package)

    return result


def get_origin_for_installed_package(
    package: apt_pkg.Package, dep_cache: apt_pkg.DepCache
) -> str:
    """
    Returns the origin for a package installed in the system.

    Technically speaking, packages don't have origins - their versions do.
    We check the available versions (installed, candidate) to determine the
    most reasonable origin for the package.
    """
    # We assume that packages we pass are installed
    if not package.current_ver:
        return ""
    available_origins = package.current_ver.file_list

    # If the installed version for a package has a single origin, it means that
    # only the local dpkg reference is there. Then, we check if there is a
    # candidate version. No candidate means we don't know anything about the
    # package. Otherwise we check for the origins of the candidate version.
    if len(available_origins) == 1:
        candidate = dep_cache.get_candidate_ver(package)
        if not candidate or package.current_ver == candidate:
            return "unknown"
        available_origins = candidate.file_list

    for origin, _ in available_origins:
        service = get_origin_information_to_service_map().get(
            (origin.origin, origin.archive), ""
        )
        if service in ESM_SERVICES:
            return service
        if origin.origin == "Ubuntu":
            return origin.component

    return "third-party"


@lru_cache(maxsize=None)
def _is_candidate_version(pkg: str, version: str) -> bool:
    """Returns True if the package version is a candidate version."""
    candidate_version = get_pkg_candidate_version(pkg, check_esm_cache=False)
    if candidate_version:
        return version == candidate_version
    return False


def get_update_status(
    service_name: str,
    ua_info: Dict[str, Any],
    version: apt_pkg.Version,
) -> str:
    """Defines the update status for a package based on the service name.

    For ESM-[Infra|Apps] packages, first checks if Pro is attached. If this is
    the case, also check for availability of the service.
    """
    if service_name in ("standard-security", "standard-updates") or (
        ua_info["attached"] and service_name in ua_info["enabled_services"]
    ):
        is_candidate = _is_candidate_version(
            version.parent_pkg.name, version.ver_str
        )
        return (
            UpdateStatus.AVAILABLE.value
            if is_candidate
            else UpdateStatus.AVAILABLE_NOT_PREFERRED.value
        )
    if not ua_info["attached"]:
        return UpdateStatus.UNATTACHED.value
    if service_name in ua_info["entitled_services"]:
        return UpdateStatus.NOT_ENABLED.value
    return UpdateStatus.UNAVAILABLE.value


def filter_updates(
    packages: List[apt_pkg.Package],
) -> DefaultDict[str, List[Tuple[apt_pkg.Version, str]]]:
    """Filters a list of packages looking for available updates.

    All versions greater than the installed one are reported, based on where
    it is provided, including ESM pockets, excluding backports.
    """
    result = defaultdict(list)

    # This esm_cache will only be usefull for the situation where
    # the user does not have the esm (infra or apps) services enabled,
    # but has be advertised about esm packages. Since those
    # sources live in a private folder, we need a different apt cache
    # to access them.
    with PreserveAptCfg(get_esm_apt_pkg_cache) as esm_cache:
        for package in packages:
            # We only care about installed packages here
            if package.current_ver:
                for version in package.version_list:
                    # Seems mypy cannot understand we can compare these :/
                    if version > package.current_ver:  # type: ignore
                        counted_as_security = False
                        for origin, _ in version.file_list:
                            service = (
                                get_origin_information_to_service_map().get(
                                    (origin.origin, origin.archive)
                                )
                            )
                            if service:
                                result[service].append((version, origin.site))
                                counted_as_security = True
                                # No need to loop through all the origins
                                break
                        # Also no need to report backports at least for now...
                        expected_origin = version.file_list[0][0]
                        if (
                            not counted_as_security
                            and "backports" not in expected_origin.archive
                        ):
                            result["standard-updates"].append(
                                (version, expected_origin.site)
                            )

                # This loop should be only used if the user does not have esm
                # (infra or apps) enabled, and it is shorter than the
                # previous one
                if package.name in esm_cache:
                    esm_package = esm_cache[package.name]
                    for version in esm_package.version_list:
                        if version > package.current_ver:  # type: ignore
                            for origin, _ in version.file_list:
                                service = get_origin_information_to_service_map().get(  # noqa: E501
                                    (origin.origin, origin.archive)
                                )
                                if service:
                                    result[service].append(
                                        (version, origin.site)
                                    )
                                    break

    return result


def get_ua_info(cfg: UAConfig) -> Dict[str, Any]:
    """Returns the Pro information based on the config object."""
    is_attached = _is_attached(cfg).is_attached
    ua_info = {
        "attached": is_attached,
        "enabled_services": [],
        "entitled_services": [],
    }  # type: Dict[str, Any]

    if is_attached:
        infra_entitlement = ESMInfraEntitlement(cfg)
        apps_entitlement = ESMAppsEntitlement(cfg)

        if apps_entitlement.contract_status() == ContractStatus.ENTITLED:
            ua_info["entitled_services"].append("esm-apps")
        if (
            apps_entitlement.application_status()[0]
            == ApplicationStatus.ENABLED
        ):
            ua_info["enabled_services"].append("esm-apps")

        if infra_entitlement.contract_status() == ContractStatus.ENTITLED:
            ua_info["entitled_services"].append("esm-infra")
        if (
            infra_entitlement.application_status()[0]
            == ApplicationStatus.ENABLED
        ):
            ua_info["enabled_services"].append("esm-infra")

    return ua_info


# Yeah Any is bad, but so is python<3.8 without TypedDict
def get_livepatch_fixed_cves() -> List[Dict[str, Any]]:
    try:
        lp_status = livepatch.status()
    except exceptions.ProcessExecutionError:
        return []

    our_kernel_version = get_kernel_info().proc_version_signature_version
    if (
        lp_status is not None
        and our_kernel_version is not None
        and our_kernel_version == lp_status.kernel
        and lp_status.livepatch is not None
        and lp_status.livepatch.state == "applied"
        and lp_status.livepatch.fixes is not None
        and len(lp_status.livepatch.fixes) > 0
    ):
        return [
            {"name": fix.name or "", "patched": fix.patched or False}
            for fix in lp_status.livepatch.fixes
        ]

    return []


def create_updates_list(
    upgradable_versions: DefaultDict[str, List[Tuple[apt_pkg.Version, str]]],
    ua_info: Dict[str, Any],
) -> List[Dict[str, Any]]:
    updates = []
    for service, version_list in upgradable_versions.items():
        for version, origin in version_list:
            status = get_update_status(service, ua_info, version)
            updates.append(
                {
                    "package": version.parent_pkg.name,
                    "version": version.ver_str,
                    "service_name": service,
                    "status": status,
                    "origin": origin,
                    "download_size": version.size,
                }
            )

    return updates


def security_status_dict(cfg: UAConfig) -> Dict[str, Any]:
    """Returns the status of security updates on a system.

    The returned dict has a 'packages' key with a list of all installed
    packages which can receive security updates, with or without ESM,
    reflecting the availability of the update based on the Pro status.

    There is also a summary with the Ubuntu Pro information and the package
    counts.
    """
    ua_info = get_ua_info(cfg)

    summary = {"ua": ua_info}  # type: Dict[str, Any]
    packages_by_origin = get_installed_packages_by_origin()

    installed_packages = packages_by_origin["all"]
    summary["num_installed_packages"] = len(installed_packages)

    upgradable_versions = filter_updates(installed_packages)
    # This version of security-status only cares about security updates
    upgradable_versions["standard-updates"] = []

    updates = create_updates_list(upgradable_versions, ua_info)

    summary["num_main_packages"] = len(packages_by_origin["main"])
    summary["num_restricted_packages"] = len(packages_by_origin["restricted"])
    summary["num_universe_packages"] = len(packages_by_origin["universe"])
    summary["num_multiverse_packages"] = len(packages_by_origin["multiverse"])
    summary["num_third_party_packages"] = len(
        packages_by_origin["third-party"]
    )
    summary["num_unknown_packages"] = len(packages_by_origin["unknown"])
    summary["num_esm_infra_packages"] = len(packages_by_origin["esm-infra"])
    summary["num_esm_apps_packages"] = len(packages_by_origin["esm-apps"])

    summary["num_esm_infra_updates"] = len(upgradable_versions["esm-infra"])
    summary["num_esm_apps_updates"] = len(upgradable_versions["esm-apps"])
    summary["num_standard_security_updates"] = len(
        upgradable_versions["standard-security"]
    )
    summary["reboot_required"] = _reboot_required(cfg).reboot_required

    return {
        "_schema_version": "0.1",
        "summary": summary,
        "packages": updates,
        "livepatch": {"fixed_cves": get_livepatch_fixed_cves()},
    }


def _print_package_summary(
    package_lists: DefaultDict[str, List[apt_pkg.Package]],
    show_items: str = "all",
    always_show: bool = False,
) -> None:
    total_packages = len(package_lists["all"])
    print(messages.SS_SUMMARY_TOTAL.format(count=total_packages))

    offset = " " * (len(str(total_packages)) + 1)

    if show_items in ("all", "esm-infra"):
        packages_mr = (
            len(package_lists["main"])
            + len(package_lists["restricted"])
            + len(package_lists["esm-infra"])
        )
        print(
            messages.SS_SUMMARY_ARCHIVE.pluralize(packages_mr).format(
                offset=offset,
                count=packages_mr,
                repository="Main/Restricted",
            )
        )

    if show_items in ("all", "esm-apps"):
        packages_um = (
            len(package_lists["universe"])
            + len(package_lists["multiverse"])
            + len(package_lists["esm-apps"])
        )
        if packages_um or always_show:
            print(
                messages.SS_SUMMARY_ARCHIVE.pluralize(packages_um).format(
                    offset=offset,
                    count=packages_um,
                    repository="Universe/Multiverse",
                )
            )

    if show_items in ("all", "third-party"):
        packages_thirdparty = len(package_lists["third-party"])
        if packages_thirdparty or always_show:
            print(
                messages.SS_SUMMARY_THIRD_PARTY.pluralize(
                    packages_thirdparty
                ).format(offset=offset, count=packages_thirdparty)
            )

    if show_items in ("all", "unknown"):
        packages_unknown = len(package_lists["unknown"])
        if packages_unknown or always_show:
            print(
                messages.SS_SUMMARY_UNAVAILABLE.pluralize(
                    packages_unknown
                ).format(
                    offset=offset,
                    count=packages_unknown,
                )
            )

    print("")


def _print_interim_release_support():
    series = get_release_info().series
    eol_date = get_distro_info(series).eol
    date = "{}/{}".format(str(eol_date.month), str(eol_date.year))
    print(messages.SS_INTERIM_SUPPORT.format(date=date))
    print("")


def _print_lts_support():
    series = get_release_info().series
    if is_supported(series):
        eol_date = get_distro_info(series).eol
        print(messages.SS_LTS_SUPPORT.format(date=str(eol_date.year)))
    else:
        print(messages.SS_NO_SECURITY_COVERAGE)


def _print_service_support(
    service: str,
    repository: str,
    service_status: ApplicationStatus,
    service_applicability: ApplicabilityStatus,
    installed_updates: int,
    available_updates: int,
    is_attached: bool,
):
    series = get_release_info().series
    eol_date_esm = get_distro_info(series).eol_esm
    if service_status == ApplicationStatus.ENABLED:
        message = messages.SS_SERVICE_ENABLED.format(
            repository=repository,
            service=service,
            year=str(eol_date_esm.year),
        )
    else:
        message = messages.SS_SERVICE_ADVERTISE.format(
            service=service,
            repository=repository,
            year=str(eol_date_esm.year),
        )

    if installed_updates:
        message += " " + messages.SS_SERVICE_ENABLED_COUNTS.pluralize(
            installed_updates
        ).format(
            updates=installed_updates,
        )

    if available_updates:
        message += " " + messages.SS_SERVICE_ADVERTISE_COUNTS.pluralize(
            available_updates
        ).format(
            updates=available_updates,
        )
    print(message)

    if (
        is_attached
        and service_status == ApplicationStatus.DISABLED
        and service_applicability == ApplicabilityStatus.APPLICABLE
    ):
        print("")
        print(messages.SS_SERVICE_COMMAND.format(service=service))

    print("")


def _print_apt_update_call():
    last_apt_update = get_apt_cache_datetime()
    if last_apt_update is None:
        print(messages.SS_UPDATE_UNKNOWN)
        print("")
        return

    now = datetime.now(timezone.utc)
    time_since_update = now - last_apt_update
    if time_since_update.days > 0:
        print(messages.SS_UPDATE_DAYS.format(days=time_since_update.days))
        print("")


def security_status(cfg: UAConfig):
    esm_infra_ent = ESMInfraEntitlement(cfg)
    esm_apps_ent = ESMAppsEntitlement(cfg)

    esm_infra_status = esm_infra_ent.application_status()[0]
    esm_infra_applicability = esm_infra_ent.applicability_status()[0]
    esm_apps_status = esm_apps_ent.application_status()[0]
    esm_apps_applicability = esm_apps_ent.applicability_status()[0]

    series = get_release_info().series
    is_lts = is_current_series_lts()
    is_attached = get_ua_info(cfg)["attached"]

    packages_by_origin = get_installed_packages_by_origin()
    security_upgradable_versions_infra = filter_updates(
        packages_by_origin["main"]
        + packages_by_origin["restricted"]
        + packages_by_origin["esm-infra"],
    )["esm-infra"]

    security_upgradable_versions_apps = filter_updates(
        packages_by_origin["universe"]
        + packages_by_origin["multiverse"]
        + packages_by_origin["esm-apps"],
    )["esm-apps"]

    _print_package_summary(packages_by_origin)

    print(messages.SS_HELP_CALL)
    print("")

    _print_apt_update_call()

    if not is_lts:
        if is_supported(series):
            _print_interim_release_support()
        print(messages.SS_NO_INTERIM_PRO_SUPPORT)
        return

    if esm_infra_status == ApplicationStatus.DISABLED:
        _print_lts_support()

    if is_attached:
        print(messages.SS_IS_ATTACHED)
    else:
        print(messages.SS_IS_NOT_ATTACHED)
    print("")

    _print_service_support(
        service="esm-infra",
        repository="Main/Restricted",
        service_status=esm_infra_status,
        service_applicability=esm_infra_applicability,
        installed_updates=len(packages_by_origin["esm-infra"]),
        available_updates=len(security_upgradable_versions_infra),
        is_attached=is_attached,
    )

    if (
        packages_by_origin["universe"]
        or packages_by_origin["multiverse"]
        or packages_by_origin["esm-apps"]
    ):
        _print_service_support(
            service="esm-apps",
            repository="Universe/Multiverse",
            service_status=esm_apps_status,
            service_applicability=esm_apps_applicability,
            installed_updates=len(packages_by_origin["esm-apps"]),
            available_updates=len(security_upgradable_versions_apps),
            is_attached=is_attached,
        )

    if not is_attached:
        print(messages.SS_LEARN_MORE)


def list_third_party_packages():
    packages_by_origin = get_installed_packages_by_origin()
    third_party_packages = packages_by_origin["third-party"]
    package_names = [package.name for package in third_party_packages]

    _print_package_summary(
        packages_by_origin, show_items="third-party", always_show=True
    )

    if third_party_packages:
        print(messages.SS_THIRD_PARTY)

        print("")
        print(messages.SS_PACKAGES_HEADER)
        print_package_list(package_names)
        print(messages.SS_SHOW_HINT.format(package=choice(package_names)))
    else:
        print(messages.SS_NO_THIRD_PARTY)


def list_unavailable_packages():
    packages_by_origin = get_installed_packages_by_origin()
    unknown_packages = packages_by_origin["unknown"]
    package_names = [package.name for package in unknown_packages]

    _print_package_summary(
        packages_by_origin, show_items="unknown", always_show=True
    )

    if unknown_packages:
        print(messages.SS_UNAVAILABLE)
        print("")

        print(messages.SS_PACKAGES_HEADER)
        print_package_list(package_names)
        print(messages.SS_SHOW_HINT.format(package=choice(package_names)))

    else:
        print(messages.SS_NO_UNAVAILABLE)


def list_esm_infra_packages(cfg):
    packages_by_origin = get_installed_packages_by_origin()
    infra_packages = packages_by_origin["esm-infra"]
    mr_packages = packages_by_origin["main"] + packages_by_origin["restricted"]

    all_infra_packages = infra_packages + mr_packages

    infra_updates = set()
    security_upgradable_versions = filter_updates(all_infra_packages)[
        "esm-infra"
    ]
    for update, _ in security_upgradable_versions:
        infra_updates.add(update.parent_pkg)

    series = get_release_info().series
    is_lts = is_current_series_lts()

    esm_infra_ent = ESMInfraEntitlement(cfg)
    esm_apps_ent = ESMAppsEntitlement(cfg)

    esm_infra_status = esm_infra_ent.application_status()[0]
    esm_infra_applicability = esm_apps_ent.applicability_status()[0]

    installed_package_names = sorted(
        [package.name for package in infra_packages]
    )
    available_package_names = sorted(
        [package.name for package in infra_updates]
    )
    remaining_package_names = sorted(
        [
            package.name
            for package in all_infra_packages
            if package.name not in installed_package_names
            and package.name not in available_package_names
        ]
    )

    _print_package_summary(
        packages_by_origin, show_items="esm-infra", always_show=True
    )

    if not is_lts:
        if is_supported(series):
            _print_interim_release_support()
        print(messages.SS_NO_INTERIM_PRO_SUPPORT)
        return

    if esm_infra_status == ApplicationStatus.DISABLED:
        _print_lts_support()
        print("")

    _print_service_support(
        service="esm-infra",
        repository="Main/Restricted",
        service_status=esm_infra_status,
        service_applicability=esm_infra_applicability,
        installed_updates=len(infra_packages),
        available_updates=len(infra_updates),
        is_attached=False,  # don't care about the `enable` message
    )
    print(messages.SS_SERVICE_HELP.format(service="esm-infra"))
    print("")

    if not is_supported(series):
        if available_package_names:
            print(messages.SS_UPDATES_AVAILABLE.format(service="esm-infra"))
            print_package_list(available_package_names)

        if installed_package_names:
            print(messages.SS_UPDATES_INSTALLED.format(service="esm-infra"))
            print_package_list(installed_package_names)

        hint_list = available_package_names or installed_package_names
        # Check names because packages may have been already listed
        if remaining_package_names:
            if hint_list:
                msg = messages.SS_FURTHER_OTHER_PACKAGES.format(
                    service="esm-infra"
                )
            else:
                msg = messages.SS_OTHER_PACKAGES.format(service="esm-infra")
            print(msg)
            print_package_list(remaining_package_names)

        if hint_list:
            print(messages.SS_SHOW_HINT.format(package=choice(hint_list)))


def list_esm_apps_packages(cfg):
    packages_by_origin = get_installed_packages_by_origin()
    apps_packages = packages_by_origin["esm-apps"]
    um_packages = (
        packages_by_origin["universe"] + packages_by_origin["multiverse"]
    )

    all_apps_packages = apps_packages + um_packages

    apps_updates = set()
    security_upgradable_versions = filter_updates(all_apps_packages)[
        "esm-apps"
    ]
    for update, _ in security_upgradable_versions:
        apps_updates.add(update.parent_pkg)

    is_lts = is_current_series_lts()

    esm_apps_ent = ESMAppsEntitlement(cfg)
    esm_apps_status = esm_apps_ent.application_status()[0]
    esm_apps_applicability = esm_apps_ent.applicability_status()[0]

    installed_package_names = sorted(
        [package.name for package in apps_packages]
    )
    available_package_names = sorted(
        [package.name for package in apps_updates]
    )
    remaining_package_names = sorted(
        [
            package.name
            for package in all_apps_packages
            if package.name not in installed_package_names
            and package.name not in available_package_names
        ]
    )

    _print_package_summary(
        packages_by_origin, show_items="esm-apps", always_show=True
    )

    if not is_lts:
        print(messages.SS_NO_INTERIM_PRO_SUPPORT)
        return

    _print_service_support(
        service="esm-apps",
        repository="Universe/Multiverse",
        service_status=esm_apps_status,
        service_applicability=esm_apps_applicability,
        installed_updates=len(apps_packages),
        available_updates=len(apps_updates),
        is_attached=False,  # don't care about the `enable` message
    )
    print(messages.SS_SERVICE_HELP.format(service="esm-apps"))
    print("")

    if all_apps_packages:
        if available_package_names:
            print(messages.SS_UPDATES_AVAILABLE.format(service="esm-apps"))
            print_package_list(available_package_names)

        if installed_package_names:
            print(messages.SS_UPDATES_INSTALLED.format(service="esm-apps"))
            print_package_list(installed_package_names)

        hint_list = available_package_names or installed_package_names

        # Check names because packages may have been already listed
        if remaining_package_names:
            if hint_list:
                msg = messages.SS_FURTHER_OTHER_PACKAGES.format(
                    service="esm-apps"
                )
            else:
                msg = messages.SS_OTHER_PACKAGES.format(service="esm-apps")
            print(msg)
            print_package_list(remaining_package_names)

        if hint_list:
            print(messages.SS_SHOW_HINT.format(package=choice(hint_list)))



import mock

from lib.esm_cache import main
from uaclient.exceptions import MissingSeriesOnOSReleaseFile
from uaclient.messages import E_MISSING_SERIES_ON_OS_RELEASE


@mock.patch("lib.esm_cache.update_esm_caches")
class TestUpdateEsmCaches:
    def test_builds_local_cache(self, m_update_caches, FakeConfig):
        main(FakeConfig())

        assert m_update_caches.call_count == 1

    @mock.patch("lib.esm_cache.LOG.error")
    def test_log_user_facing_exception(
        self, m_esm_cache_log_err, m_update_caches, FakeConfig
    ):
        expected_exception = MissingSeriesOnOSReleaseFile(version="version")
        m_update_caches.side_effect = expected_exception
        main(cfg=None)
        expected_log_args = [
            mock.call(
                "Error updating the cache: %s",
                E_MISSING_SERIES_ON_OS_RELEASE.format(version="version").msg,
            )
        ]

        assert expected_log_args == m_esm_cache_log_err.call_args_list

    def test_log_exception(self, m_update_caches, caplog_text, FakeConfig):
        expected_msg = "unexpected exception"
        expected_exception = Exception(expected_msg)
        m_update_caches.side_effect = expected_exception
        main(cfg=FakeConfig())
        log = caplog_text()

        assert expected_msg in log



from argparse import Namespace

import mock
import pytest

from uaclient import exceptions, messages
from uaclient.api.u.pro.services.dependencies.v1 import (
    DependenciesResult,
    ServiceWithDependencies,
    ServiceWithReason,
)
from uaclient.api.u.pro.services.enable.v1 import EnableOptions, EnableResult
from uaclient.api.u.pro.status.enabled_services.v1 import (
    EnabledService,
    EnabledServicesResult,
)
from uaclient.api.u.pro.status.is_attached.v1 import IsAttachedResult
from uaclient.cli.enable import (
    _enable_landscape,
    _enable_one_service,
    _EnableOneServiceResult,
    _print_json_output,
    enable_command,
    prompt_for_dependency_handling,
)
from uaclient.testing.helpers import does_not_raise


class TestActionEnable:
    @pytest.mark.parametrize(
        [
            "is_attached",
            "args",
            "kwargs",
            "refresh_side_effect",
            "valid_entitlement_names",
            "enabled_services",
            "dependencies",
            "entitlements_for_enabling",
            "enable_one_service_side_effect",
            "expected_enable_one_service_calls",
            "expected_print_json_output_calls",
            "expected_raises",
        ],
        (
            # assume-yes required for json output
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=True,
                ),
                Namespace(
                    format="json",
                    variant="",
                    access_only=False,
                    assume_yes=False,
                    auto=False,
                ),
                {},
                None,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                pytest.raises(exceptions.CLIJSONFormatRequireAssumeYes),
            ),
            # variant + access-only not allowed
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=True,
                ),
                Namespace(
                    format="cli",
                    variant="variant",
                    access_only=True,
                    assume_yes=False,
                    auto=False,
                ),
                {},
                None,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                pytest.raises(exceptions.InvalidOptionCombination),
            ),
            # variant + auto not allowed
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=True,
                ),
                Namespace(
                    format="cli",
                    variant="variant",
                    access_only=False,
                    assume_yes=False,
                    auto=True,
                ),
                {},
                None,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                pytest.raises(exceptions.InvalidOptionCombination),
            ),
            # access_only + auto not allowed
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=True,
                ),
                Namespace(
                    format="cli",
                    variant="",
                    access_only=True,
                    assume_yes=False,
                    auto=True,
                ),
                {},
                None,
                None,
                None,
                None,
                None,
                None,
                [],
                [],
                pytest.raises(exceptions.InvalidOptionCombination),
            ),
            # contract not valid
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=False,
                ),
                Namespace(
                    format="cli",
                    variant="",
                    access_only=False,
                    assume_yes=False,
                    auto=False,
                ),
                {},
                None,
                None,
                None,
                None,
                None,
                None,
                [],
                [
                    mock.call(
                        False,
                        {"_schema_version": "0.1", "needs_reboot": False},
                        [],
                        [],
                        [
                            {
                                "type": "system",
                                "message": messages.E_CONTRACT_EXPIRED.msg,
                                "message_code": messages.E_CONTRACT_EXPIRED.name,  # noqa: E501
                            }
                        ],
                        [],
                        success=False,
                    )
                ],
                does_not_raise(),
            ),
            # contract not valid, json output
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=False,
                ),
                Namespace(
                    format="json",
                    variant="",
                    access_only=False,
                    assume_yes=True,
                    auto=False,
                ),
                {},
                None,
                None,
                None,
                None,
                None,
                None,
                [],
                [
                    mock.call(
                        True,
                        {"_schema_version": "0.1", "needs_reboot": False},
                        [],
                        [],
                        [
                            {
                                "type": "system",
                                "message": messages.E_CONTRACT_EXPIRED.msg,
                                "message_code": messages.E_CONTRACT_EXPIRED.name,  # noqa: E501
                            }
                        ],
                        [],
                        success=False,
                    )
                ],
                does_not_raise(),
            ),
            # contract not valid, json output, contract refresh warning
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=False,
                ),
                Namespace(
                    format="json",
                    variant="",
                    access_only=False,
                    assume_yes=True,
                    auto=False,
                ),
                {},
                exceptions.ConnectivityError(cause=Exception(), url=""),
                None,
                None,
                None,
                None,
                None,
                [],
                [
                    mock.call(
                        True,
                        {"_schema_version": "0.1", "needs_reboot": False},
                        [],
                        [],
                        [
                            {
                                "type": "system",
                                "message": messages.E_CONTRACT_EXPIRED.msg,
                                "message_code": messages.E_CONTRACT_EXPIRED.name,  # noqa: E501
                            }
                        ],
                        [
                            {
                                "type": "system",
                                "message": messages.E_REFRESH_CONTRACT_FAILURE.msg,  # noqa: E501
                                "message_code": messages.E_REFRESH_CONTRACT_FAILURE.name,  # noqa: E501
                            }
                        ],
                        success=False,
                    )
                ],
                does_not_raise(),
            ),
            # success multiple services, one needs reboot
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=True,
                ),
                Namespace(
                    service=["one", "two", "three"],
                    format="json",
                    variant="",
                    access_only=False,
                    assume_yes=True,
                    auto=False,
                ),
                {},
                None,
                (["one", "two", "three"], []),
                EnabledServicesResult(enabled_services=[]),
                DependenciesResult(services=mock.sentinel.dependencies),
                ["three", "two", "one"],
                [
                    _EnableOneServiceResult(
                        success=True, needs_reboot=False, error=None
                    ),
                    _EnableOneServiceResult(
                        success=True, needs_reboot=True, error=None
                    ),
                    _EnableOneServiceResult(
                        success=True, needs_reboot=False, error=None
                    ),
                ],
                [
                    mock.call(
                        mock.ANY,
                        "three",
                        "",
                        False,
                        True,
                        True,
                        None,
                        [],
                        mock.sentinel.dependencies,
                    ),
                    mock.call(
                        mock.ANY,
                        "two",
                        "",
                        False,
                        True,
                        True,
                        None,
                        [],
                        mock.sentinel.dependencies,
                    ),
                    mock.call(
                        mock.ANY,
                        "one",
                        "",
                        False,
                        True,
                        True,
                        None,
                        [],
                        mock.sentinel.dependencies,
                    ),
                ],
                [
                    mock.call(
                        True,
                        {"_schema_version": "0.1", "needs_reboot": True},
                        ["three", "two", "one"],
                        [],
                        [],
                        [],
                        success=True,
                    )
                ],
                does_not_raise(),
            ),
            # some services not found
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=True,
                ),
                Namespace(
                    service=["one", "two", "three"],
                    format="json",
                    variant="",
                    access_only=False,
                    assume_yes=True,
                    auto=False,
                ),
                {},
                None,
                (["two"], ["one", "three"]),
                EnabledServicesResult(enabled_services=[]),
                DependenciesResult(services=mock.sentinel.dependencies),
                ["two"],
                [
                    _EnableOneServiceResult(
                        success=True, needs_reboot=False, error=None
                    ),
                ],
                [
                    mock.call(
                        mock.ANY,
                        "two",
                        "",
                        False,
                        True,
                        True,
                        None,
                        [],
                        mock.sentinel.dependencies,
                    ),
                ],
                [
                    mock.call(
                        True,
                        {"_schema_version": "0.1", "needs_reboot": False},
                        ["two"],
                        ["one", "three"],
                        [
                            {
                                "type": "system",
                                "service": None,
                                "message": mock.ANY,
                                "message_code": "invalid-service-or-failure",
                                "additional_info": {
                                    "operation": "enable",
                                    "invalid_service": "one, three",
                                    "service_msg": mock.ANY,
                                },
                            }
                        ],
                        [],
                        success=False,
                    )
                ],
                does_not_raise(),
            ),
            # one success, one fail, one not found
            (
                IsAttachedResult(
                    is_attached=True,
                    contract_status="",
                    contract_remaining_days=100,
                    is_attached_and_contract_valid=True,
                ),
                Namespace(
                    service=["one", "two", "three"],
                    format="json",
                    variant="",
                    access_only=False,
                    assume_yes=True,
                    auto=False,
                ),
                {},
                None,
                (["one", "two"], ["three"]),
                EnabledServicesResult(enabled_services=[]),
                DependenciesResult(services=mock.sentinel.dependencies),
                ["two", "one"],
                [
                    _EnableOneServiceResult(
                        success=False,
                        needs_reboot=False,
                        error={"test": "error"},
                    ),
                    _EnableOneServiceResult(
                        success=True, needs_reboot=False, error=None
                    ),
                ],
                [
                    mock.call(
                        mock.ANY,
                        "two",
                        "",
                        False,
                        True,
                        True,
                        None,
                        [],
                        mock.sentinel.dependencies,
                    ),
                    mock.call(
                        mock.ANY,
                        "one",
                        "",
                        False,
                        True,
                        True,
                        None,
                        [],
                        mock.sentinel.dependencies,
                    ),
                ],
                [
                    mock.call(
                        True,
                        {"_schema_version": "0.1", "needs_reboot": False},
                        ["one"],
                        ["two", "three"],
                        [
                            {"test": "error"},
                            {
                                "type": "system",
                                "service": None,
                                "message": mock.ANY,
                                "message_code": "invalid-service-or-failure",
                                "additional_info": {
                                    "operation": "enable",
                                    "invalid_service": "three",
                                    "service_msg": mock.ANY,
                                },
                            },
                        ],
                        [],
                        success=False,
                    )
                ],
                does_not_raise(),
            ),
        ),
    )
    @mock.patch("uaclient.contract.UAContractClient.update_activity_token")
    @mock.patch("uaclient.cli.enable._enable_one_service")
    @mock.patch("uaclient.entitlements.order_entitlements_for_enabling")
    @mock.patch("uaclient.cli.enable._dependencies")
    @mock.patch("uaclient.cli.enable._enabled_services")
    @mock.patch("uaclient.entitlements.get_valid_entitlement_names")
    @mock.patch("uaclient.cli.enable._print_json_output")
    @mock.patch("uaclient.contract.refresh")
    @mock.patch("uaclient.cli.cli_util.create_interactive_only_print_function")
    @mock.patch("uaclient.util.we_are_currently_root", return_value=True)
    @mock.patch("uaclient.cli.enable._is_attached")
    def test_action_enable(
        self,
        m_is_attached,
        m_we_are_currently_root,
        m_create_interactive_only_print_function,
        m_refresh,
        m_print_json_output,
        m_get_valid_entitlement_names,
        m_enabled_services,
        m_dependencies,
        m_order_entitlements_for_enabling,
        m_enable_one_service,
        m_update_activity_token,
        is_attached,
        args,
        kwargs,
        refresh_side_effect,
        valid_entitlement_names,
        enabled_services,
        dependencies,
        entitlements_for_enabling,
        enable_one_service_side_effect,
        expected_enable_one_service_calls,
        expected_print_json_output_calls,
        expected_raises,
        FakeConfig,
        fake_machine_token_file,
    ):
        m_is_attached.return_value = is_attached
        m_refresh.side_effect = refresh_side_effect
        m_get_valid_entitlement_names.return_value = valid_entitlement_names
        m_enabled_services.return_value = enabled_services
        m_dependencies.return_value = dependencies
        m_order_entitlements_for_enabling.return_value = (
            entitlements_for_enabling
        )
        m_enable_one_service.side_effect = enable_one_service_side_effect
        fake_machine_token_file.attached = True

        with expected_raises:
            enable_command.action(args, cfg=FakeConfig(), **kwargs)

        assert (
            expected_enable_one_service_calls
            == m_enable_one_service.call_args_list
        )
        assert (
            expected_print_json_output_calls
            == m_print_json_output.call_args_list
        )

    @pytest.mark.parametrize(
        [
            "kwargs",
            "prompt_for_dependency_handling_side_effect",
            "enable_landscape_result",
            "enable_result",
            "expected_prompt_for_dependency_handling_calls",
            "expected_enable_landscape_calls",
            "expected_enable_calls",
            "expected_status_calls",
            "expected_result",
        ],
        (
            # already enabled
            (
                {
                    "ent_name": "one",
                    "variant": "",
                    "access_only": False,
                    "assume_yes": False,
                    "json_output": False,
                    "extra_args": None,
                    "enabled_services": [EnabledService(name="one")],
                    "all_dependencies": [],
                },
                None,
                None,
                None,
                [],
                [],
                [],
                [],
                _EnableOneServiceResult(
                    success=False,
                    needs_reboot=False,
                    error={
                        "type": "service",
                        "service": "one",
                        "message": messages.ALREADY_ENABLED.format(
                            title=mock.sentinel.ent_title
                        ).msg,
                        "message_code": "service-already-enabled",
                    },
                ),
            ),
            # prompt denied fails
            (
                {
                    "ent_name": "one",
                    "variant": "",
                    "access_only": False,
                    "assume_yes": False,
                    "json_output": False,
                    "extra_args": None,
                    "enabled_services": [],
                    "all_dependencies": [],
                },
                [exceptions.PromptDeniedError()],
                None,
                None,
                [
                    mock.call(
                        mock.ANY,
                        "one",
                        [],
                        [],
                        called_name="one",
                        variant="",
                        service_title=mock.sentinel.ent_title,
                    )
                ],
                [],
                [],
                [],
                _EnableOneServiceResult(
                    success=False, needs_reboot=False, error=None
                ),
            ),
            # landscape
            (
                {
                    "ent_name": "landscape",
                    "variant": "",
                    "access_only": mock.sentinel.access_only,
                    "assume_yes": mock.sentinel.assume_yes,
                    "json_output": False,
                    "extra_args": mock.sentinel.extra_args,
                    "enabled_services": [],
                    "all_dependencies": [],
                },
                None,
                EnableResult(
                    enabled=["landscape"],
                    disabled=[],
                    reboot_required=mock.sentinel.reboot,
                    messages=[],
                ),
                None,
                [],
                [
                    mock.call(
                        mock.ANY,
                        mock.sentinel.access_only,
                        extra_args=mock.sentinel.extra_args,
                        progress_object=mock.sentinel.cli_progress,
                    )
                ],
                [],
                [mock.call(cfg=mock.ANY)],
                _EnableOneServiceResult(
                    success=True, needs_reboot=mock.sentinel.reboot, error=None
                ),
            ),
            # non-landscape
            (
                {
                    "ent_name": "one",
                    "variant": mock.sentinel.variant,
                    "access_only": mock.sentinel.access_only,
                    "assume_yes": mock.sentinel.assume_yes,
                    "json_output": False,
                    "extra_args": mock.sentinel.extra_args,
                    "enabled_services": [],
                    "all_dependencies": [],
                },
                None,
                None,
                EnableResult(
                    enabled=["one"],
                    disabled=[],
                    reboot_required=mock.sentinel.reboot,
                    messages=[],
                ),
                [],
                [],
                [
                    mock.call(
                        EnableOptions(
                            service="one",
                            variant=mock.sentinel.variant,
                            access_only=mock.sentinel.access_only,
                        ),
                        mock.ANY,
                        progress_object=mock.sentinel.cli_progress,
                    )
                ],
                [mock.call(cfg=mock.ANY)],
                _EnableOneServiceResult(
                    success=True, needs_reboot=mock.sentinel.reboot, error=None
                ),
            ),
            # json output
            (
                {
                    "ent_name": "one",
                    "variant": mock.sentinel.variant,
                    "access_only": mock.sentinel.access_only,
                    "assume_yes": mock.sentinel.assume_yes,
                    "json_output": True,
                    "extra_args": mock.sentinel.extra_args,
                    "enabled_services": [],
                    "all_dependencies": [],
                },
                None,
                None,
                EnableResult(
                    enabled=["one"],
                    disabled=[],
                    reboot_required=mock.sentinel.reboot,
                    messages=[],
                ),
                [],
                [],
                [
                    mock.call(
                        EnableOptions(
                            service="one",
                            variant=mock.sentinel.variant,
                            access_only=mock.sentinel.access_only,
                        ),
                        mock.ANY,
                        progress_object=None,
                    )
                ],
                [mock.call(cfg=mock.ANY)],
                _EnableOneServiceResult(
                    success=True, needs_reboot=mock.sentinel.reboot, error=None
                ),
            ),
        ),
    )
    @mock.patch("uaclient.status.status")
    @mock.patch("uaclient.cli.enable._enable")
    @mock.patch("uaclient.cli.enable._enable_landscape")
    @mock.patch("uaclient.cli.enable.cli_util.CLIEnableDisableProgress")
    @mock.patch("uaclient.cli.enable.prompt_for_dependency_handling")
    @mock.patch("uaclient.cli.enable.entitlements.entitlement_factory")
    @mock.patch("uaclient.cli.cli_util.create_interactive_only_print_function")
    def test_enable_one_service(
        self,
        m_create_interactive_only_print_function,
        m_entitlement_factory,
        m_prompt_for_dependency_handling,
        m_progress_class,
        m_enable_landscape,
        m_enable,
        m_status,
        kwargs,
        prompt_for_dependency_handling_side_effect,
        enable_landscape_result,
        enable_result,
        expected_prompt_for_dependency_handling_calls,
        expected_enable_landscape_calls,
        expected_enable_calls,
        expected_status_calls,
        expected_result,
        FakeConfig,
    ):
        mock_ent = mock.MagicMock()
        m_entitlement_factory.return_value = mock_ent
        mock_ent.name = kwargs.get("ent_name")
        mock_ent.title = mock.sentinel.ent_title
        m_prompt_for_dependency_handling.side_effect = (
            prompt_for_dependency_handling_side_effect
        )
        m_progress_class.return_value = mock.sentinel.cli_progress
        m_enable_landscape.return_value = enable_landscape_result
        m_enable.return_value = enable_result

        assert expected_result == _enable_one_service(FakeConfig(), **kwargs)

        assert (
            expected_prompt_for_dependency_handling_calls
            == m_prompt_for_dependency_handling.call_args_list
        )
        assert (
            expected_enable_landscape_calls
            == m_enable_landscape.call_args_list
        )
        assert expected_enable_calls == m_enable.call_args_list
        assert expected_status_calls == m_status.call_args_list

    @pytest.mark.parametrize(
        [
            "enable_side_effect",
            "expected_raises",
            "expected_result",
        ],
        (
            (
                exceptions.LandscapeConfigFailed(),
                pytest.raises(exceptions.LandscapeConfigFailed),
                None,
            ),
            (
                [(False, None)],
                pytest.raises(exceptions.EntitlementNotEnabledError),
                None,
            ),
            (
                [(True, None)],
                does_not_raise(),
                EnableResult(
                    enabled=["landscape"],
                    disabled=[],
                    reboot_required=False,
                    messages=[],
                ),
            ),
        ),
    )
    @mock.patch("uaclient.cli.enable.lock.RetryLock")
    @mock.patch("uaclient.cli.enable.entitlements.LandscapeEntitlement")
    def test_enable_landscape(
        self,
        m_landscape_entitlement,
        m_lock,
        enable_side_effect,
        expected_raises,
        expected_result,
        FakeConfig,
    ):
        m_enable = m_landscape_entitlement.return_value.enable
        m_enable.side_effect = enable_side_effect
        with expected_raises:
            assert expected_result == _enable_landscape(
                FakeConfig,
                mock.sentinel.access_only,
                mock.sentinel.extra_args,
                None,
            )
        assert [
            mock.call(
                mock.ANY,
                called_name="landscape",
                access_only=mock.sentinel.access_only,
                extra_args=mock.sentinel.extra_args,
            )
        ] == m_landscape_entitlement.call_args_list

    @pytest.mark.parametrize(
        [
            "json_output",
            "expected_print_calls",
        ],
        (
            (True, [mock.call(mock.ANY)]),
            (False, []),
        ),
    )
    @mock.patch("builtins.print")
    def test_print_json_output(
        self, m_print, json_output, expected_print_calls
    ):
        _print_json_output(json_output, {}, [], [], [], [], True)
        assert expected_print_calls == m_print.call_args_list


class TestPromptForDependencyHandling:
    @pytest.mark.parametrize(
        [
            "service",
            "all_dependencies",
            "enabled_services",
            "called_name",
            "service_title",
            "variant",
            "cfg_block_disable_on_enable",
            "prompt_side_effects",
            "expected_prompts",
            "expected_raise",
        ],
        [
            # no dependencies
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one", incompatible_with=[], depends_on=[]
                    )
                ],
                [],
                "one",
                "One",
                "",
                False,
                [],
                [],
                does_not_raise(),
            ),
            # incompatible with "two", but two not enabled
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            )
                        ],
                        depends_on=[],
                    )
                ],
                [],
                "one",
                "One",
                "",
                False,
                [],
                [],
                does_not_raise(),
            ),
            # incompatible with "two", two enabled, successful prompt
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            )
                        ],
                        depends_on=[],
                    )
                ],
                [EnabledService(name="two")],
                "one",
                "One",
                "",
                False,
                [True],
                [mock.call(msg=mock.ANY)],
                does_not_raise(),
            ),
            # incompatible with "two", two enabled, cfg denies
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            )
                        ],
                        depends_on=[],
                    )
                ],
                [EnabledService(name="two")],
                "one",
                "One",
                "",
                True,
                [],
                [],
                pytest.raises(exceptions.IncompatibleServiceStopsEnable),
            ),
            # incompatible with "two", two enabled, denied prompt
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            )
                        ],
                        depends_on=[],
                    )
                ],
                [EnabledService(name="two")],
                "one",
                "One",
                "",
                False,
                [False],
                [mock.call(msg=mock.ANY)],
                pytest.raises(exceptions.IncompatibleServiceStopsEnable),
            ),
            # incompatible with "two" and "three", three enabled, success
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            ),
                            ServiceWithReason(
                                name="three", reason=mock.MagicMock()
                            ),
                        ],
                        depends_on=[],
                    )
                ],
                [EnabledService(name="three")],
                "one",
                "One",
                "",
                False,
                [True],
                [mock.call(msg=mock.ANY)],
                does_not_raise(),
            ),
            # depends on "two", but two already enabled
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            )
                        ],
                    )
                ],
                [EnabledService(name="two")],
                "one",
                "One",
                "",
                False,
                [],
                [],
                does_not_raise(),
            ),
            # depends on "two", two not enabled, successful prompt
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            )
                        ],
                    )
                ],
                [],
                "one",
                "One",
                "",
                False,
                [True],
                [mock.call(msg=mock.ANY)],
                does_not_raise(),
            ),
            # depends on "two", two not enabled, denied prompt
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            )
                        ],
                    )
                ],
                [],
                "one",
                "One",
                "",
                False,
                [False],
                [mock.call(msg=mock.ANY)],
                pytest.raises(exceptions.RequiredServiceStopsEnable),
            ),
            # depends on "two" and "three", three not enabled, success prompt
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[],
                        depends_on=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            ),
                            ServiceWithReason(
                                name="three", reason=mock.MagicMock()
                            ),
                        ],
                    )
                ],
                [EnabledService(name="two")],
                "one",
                "One",
                "",
                False,
                [True],
                [mock.call(msg=mock.ANY)],
                does_not_raise(),
            ),
            # a variant is enabled, second variant is being enabled
            (
                "two",
                [],
                [
                    EnabledService(
                        name="two", variant_enabled=True, variant_name="one"
                    )
                ],
                "two",
                "Two",
                "two",
                False,
                [False],
                [mock.call(msg=mock.ANY)],
                pytest.raises(exceptions.IncompatibleServiceStopsEnable),
            ),
            # lots of stuff
            (
                "one",
                [
                    ServiceWithDependencies(
                        name="one",
                        incompatible_with=[
                            ServiceWithReason(
                                name="two", reason=mock.MagicMock()
                            ),
                            ServiceWithReason(
                                name="three", reason=mock.MagicMock()
                            ),
                            ServiceWithReason(
                                name="four", reason=mock.MagicMock()
                            ),
                        ],
                        depends_on=[
                            ServiceWithReason(
                                name="five", reason=mock.MagicMock()
                            ),
                            ServiceWithReason(
                                name="six", reason=mock.MagicMock()
                            ),
                            ServiceWithReason(
                                name="seven", reason=mock.MagicMock()
                            ),
                        ],
                    )
                ],
                [
                    EnabledService(name="two"),
                    EnabledService(name="four"),
                    EnabledService(name="six"),
                ],
                "one",
                "One",
                "",
                False,
                [True, True, True, True],
                [
                    mock.call(msg=mock.ANY),
                    mock.call(msg=mock.ANY),
                    mock.call(msg=mock.ANY),
                    mock.call(msg=mock.ANY),
                ],
                does_not_raise(),
            ),
        ],
    )
    @mock.patch("uaclient.entitlements.get_title")
    @mock.patch("uaclient.util.prompt_for_confirmation")
    @mock.patch("uaclient.util.is_config_value_true")
    def test_prompt_for_dependency_handling(
        self,
        m_is_config_value_true,
        m_prompt_for_confirmation,
        m_entitlement_get_title,
        service,
        all_dependencies,
        enabled_services,
        called_name,
        service_title,
        variant,
        cfg_block_disable_on_enable,
        prompt_side_effects,
        expected_prompts,
        expected_raise,
        FakeConfig,
    ):
        m_entitlement_get_title.side_effect = (
            lambda cfg, name, variant=variant: name.title()
        )
        m_is_config_value_true.return_value = cfg_block_disable_on_enable
        m_prompt_for_confirmation.side_effect = prompt_side_effects

        with expected_raise:
            prompt_for_dependency_handling(
                FakeConfig(),
                service,
                all_dependencies,
                enabled_services,
                called_name,
                variant,
                service_title,
            )

        assert expected_prompts == m_prompt_for_confirmation.call_args_list



import copy
import datetime
import string

import mock
import pytest

from uaclient import messages, status
from uaclient.entitlements import (
    ENTITLEMENT_CLASSES,
    entitlement_factory,
    valid_services,
)
from uaclient.entitlements.base import EntitlementWithMessage
from uaclient.entitlements.entitlement_status import (
    ApplicationStatus,
    ContractStatus,
    UserFacingConfigStatus,
    UserFacingStatus,
)
from uaclient.entitlements.fips import FIPSEntitlement
from uaclient.entitlements.ros import ROSEntitlement
from uaclient.entitlements.tests.test_base import ConcreteTestEntitlement
from uaclient.files.notices import Notice, NoticesManager
from uaclient.files.user_config_file import UserConfigData
from uaclient.status import (
    DEFAULT_STATUS,
    TxtColor,
    colorize_commands,
    format_tabular,
)

DEFAULT_CFG_STATUS = {
    "execution_status": DEFAULT_STATUS["execution_status"],
    "execution_details": DEFAULT_STATUS["execution_details"],
}
M_PATH = "uaclient.entitlements."


@pytest.fixture
def all_resources_available(FakeConfig):
    resources = [
        {"name": name, "available": True}
        for name in valid_services(cfg=FakeConfig())
    ]
    return resources


@pytest.fixture(params=[True, False])
def status_dict_attached(request):
    status = DEFAULT_STATUS.copy()

    # The following are required so we don't get an "unattached" error
    status["attached"] = True
    status["expires"] = datetime.datetime.now(datetime.timezone.utc)
    status["account"] = {"name": ""}
    status["contract"] = {
        "name": "",
        "tech_support_level": UserFacingStatus.INAPPLICABLE.value,
    }

    if request.param:
        status["account"]["name"] = "account"
        status["contract"]["name"] = "subscription"

    return status


@pytest.fixture
def status_dict_unattached():
    status = DEFAULT_STATUS.copy()

    status["services"] = [
        {
            "name": "cc-eal",
            "description": "Common Criteria EAL2 Provisioning Packages",
            "available": "no",
        }
    ]

    return status


class TestColorizeCommands:
    @pytest.mark.parametrize(
        "commands,expected",
        [
            (
                [
                    ["apt", "update"],
                    ["apt", "install", "--only-upgrade", "-y", "pkg"],
                ],
                TxtColor.DISABLEGREY
                + "{ apt update && apt install --only-upgrade -y pkg }"
                + TxtColor.ENDC,
            ),
            (
                [
                    ["apt", "update"],
                    [
                        "apt",
                        "install",
                        "--only-upgrade",
                        "-y",
                        "longpackagename1",
                        "longpackagename2",
                        "longpackagename3",
                        "longpackagename4",
                        "longpackagename5",
                        "longpackagename6",
                        "longpackagename7",
                        "longpackagename8",
                        "longpackagename9",
                        "longpackagename10",
                    ],
                ],
                TxtColor.DISABLEGREY
                + "{\n"
                + "  apt update && apt install --only-upgrade -y longpackagename1 \\\n"  # noqa: E501
                + "  longpackagename2 longpackagename3 longpackagename4 longpackagename5 \\\n"  # noqa: E501
                + "  longpackagename6 longpackagename7 longpackagename8 longpackagename9 \\\n"  # noqa: E501
                + "  longpackagename10"
                + "\n}"
                + TxtColor.ENDC,
            ),
        ],
    )
    def test_colorize_commands(self, commands, expected):
        assert colorize_commands(commands) == expected


@mock.patch("uaclient.livepatch.on_supported_kernel", return_value=None)
class TestFormatTabular:
    @pytest.mark.parametrize(
        "support_level,expected_colour,istty",
        [
            ("n/a", TxtColor.DISABLEGREY, True),
            ("essential", TxtColor.OKGREEN, True),
            ("standard", TxtColor.OKGREEN, True),
            ("advanced", TxtColor.OKGREEN, True),
            ("something else", None, True),
            ("n/a", TxtColor.DISABLEGREY, True),
            ("essential", None, False),
            ("standard", None, False),
            ("advanced", None, False),
            ("something else", None, False),
            ("n/a", None, False),
        ],
    )
    @mock.patch("sys.stdout.isatty")
    def test_support_colouring(
        self,
        m_isatty,
        m_on_supported_kernel,
        support_level,
        expected_colour,
        istty,
        status_dict_attached,
    ):
        status_dict_attached["contract"]["tech_support_level"] = support_level

        m_isatty.return_value = istty
        tabular_output = format_tabular(status_dict_attached)

        expected_string = "Technical support level: {}".format(
            support_level
            if not expected_colour
            else expected_colour + support_level + TxtColor.ENDC
        )
        assert expected_string in tabular_output

    @pytest.mark.parametrize("origin", ["free", "not-free"])
    def test_header_alignment(
        self, m_on_supported_kernel, origin, status_dict_attached
    ):
        status_dict_attached["origin"] = origin
        tabular_output = format_tabular(status_dict_attached)
        colon_idx = None
        for line in tabular_output.splitlines():
            if ":" not in line or "Enable services" in line:
                # This isn't a header line
                continue
            if colon_idx is None:
                # This is the first header line, record where the colon is
                colon_idx = line.index(":")
                continue
            # Ensure that the colon in this line is aligned with previous ones
            assert line.index(":") == colon_idx

    @pytest.mark.parametrize(
        "origin,expected_headers",
        [
            ("free", ()),
            ("not-free", ("Valid until", "Technical support level")),
        ],
    )
    def test_correct_header_keys_included(
        self,
        m_on_supported_kernel,
        origin,
        expected_headers,
        status_dict_attached,
    ):
        status_dict_attached["origin"] = origin

        if status_dict_attached["contract"].get("name"):
            expected_headers = ("Subscription",) + expected_headers
        if status_dict_attached["account"].get("name"):
            expected_headers = ("Account",) + expected_headers

        tabular_output = format_tabular(status_dict_attached)

        headers = [
            line.split(":")[0].strip()
            for line in tabular_output.splitlines()
            if ":" in line and "Enable services" not in line
        ]
        assert list(expected_headers) == headers

    def test_correct_unattached_column_alignment(
        self, m_on_supported_kernel, status_dict_unattached
    ):
        tabular_output = format_tabular(status_dict_unattached)
        [header, eal_service_line] = [
            line
            for line in tabular_output.splitlines()
            if "eal" in line or "AVAILABLE" in line
        ]
        printable_eal_line = "".join(
            filter(lambda x: x in string.printable, eal_service_line)
        )
        assert header.find("AVAILABLE") == printable_eal_line.find("no")
        assert header.find("DESCRIPTION") == printable_eal_line.find("Common")

    @pytest.mark.parametrize("attached", [True, False])
    def test_no_leading_newline(
        self,
        m_on_supported_kernel,
        attached,
        status_dict_attached,
        status_dict_unattached,
    ):
        if attached:
            status_dict = status_dict_attached
        else:
            status_dict = status_dict_unattached

        assert not format_tabular(status_dict).startswith("\n")

    @pytest.mark.parametrize(
        "description_override, uf_status, uf_descr",
        (
            ("", "n/a", "Common Criteria EAL2 default descr"),
            ("Custom descr", "n/a", "Custom descr"),
            ("Custom call to action", "enabled", "Custom call to action"),
        ),
    )
    def test_custom_descr(
        self,
        m_on_supported_kernel,
        description_override,
        uf_status,
        uf_descr,
        status_dict_attached,
    ):
        """Services can provide a custom call to action if present."""
        default_descr = "Common Criteria EAL2 default descr"
        status_dict_attached["services"] = [
            {
                "name": "cc-eal",
                "description": default_descr,
                "available": "no",
                "status": uf_status,
                "entitled": True,
                "description_override": description_override,
            }
        ]
        if not description_override:
            # Remove key to test upgrade path from older ua-tools
            status_dict_attached["services"][0].pop("description_override")
        assert uf_descr in format_tabular(status_dict_attached)


@pytest.fixture
def esm_desc(FakeConfig):
    return entitlement_factory(cfg=FakeConfig(), name="esm-infra").description


@pytest.fixture
def ros_desc(FakeConfig):
    return entitlement_factory(cfg=FakeConfig(), name="ros").description


@mock.patch("uaclient.livepatch.on_supported_kernel", return_value=None)
@mock.patch("uaclient.files.notices.NoticesManager.remove")
@mock.patch("uaclient.system.should_reboot", return_value=False)
class TestStatus:
    @pytest.mark.parametrize("show_all", (True, False))
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.status.get_available_resources")
    def test_root_unattached(
        self,
        m_get_available_resources,
        _m_status_cache_file,
        _m_should_reboot,
        _m_remove_notice,
        m_on_supported_kernel,
        ros_desc,
        esm_desc,
        show_all,
        FakeConfig,
    ):
        """Test we get the correct status dict when unattached"""
        if show_all:
            expected_services = [
                {
                    "available": "yes",
                    "name": "esm-infra",
                    "description": esm_desc,
                    "description_override": None,
                },
                {
                    "available": "no",
                    "name": "ros",
                    "description": ros_desc,
                    "description_override": None,
                },
            ]
        else:
            expected_services = [
                {
                    "available": "yes",
                    "name": "esm-infra",
                    "description": esm_desc,
                    "description_override": None,
                }
            ]
        cfg = FakeConfig()
        m_get_available_resources.return_value = [
            {"name": "esm-infra", "available": True},
            {"name": "ros", "available": False},
        ]
        expected = copy.deepcopy(DEFAULT_STATUS)
        expected["version"] = mock.ANY
        expected["services"] = expected_services
        with mock.patch(
            "uaclient.status._get_config_status"
        ) as m_get_cfg_status:
            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
            assert expected == status.status(cfg=cfg, show_all=show_all)

    @pytest.mark.parametrize(
        "avail_res,entitled_res,uf_entitled,uf_status",
        (
            (  # Empty lists means UNENTITLED and UNAVAILABLE
                [],
                [],
                ContractStatus.UNENTITLED.value,
                UserFacingStatus.UNAVAILABLE.value,
            ),
            (  # available == False means UNAVAILABLE
                [{"name": "livepatch", "available": False}],
                [],
                ContractStatus.UNENTITLED.value,
                UserFacingStatus.UNAVAILABLE.value,
            ),
            (  # available == True but unentitled means UNAVAILABLE
                [{"name": "livepatch", "available": True}],
                [],
                ContractStatus.UNENTITLED.value,
                UserFacingStatus.UNAVAILABLE.value,
            ),
            (  # available == False and entitled means INAPPLICABLE
                [{"name": "livepatch", "available": False}],
                [{"type": "livepatch", "entitled": True}],
                ContractStatus.ENTITLED.value,
                UserFacingStatus.INAPPLICABLE.value,
            ),
        ),
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch(
        M_PATH + "livepatch.LivepatchEntitlement.application_status",
        return_value=(ApplicationStatus.DISABLED, ""),
    )
    @mock.patch("uaclient.status.get_available_resources")
    def test_root_attached(
        self,
        m_get_avail_resources,
        _m_livepatch_status,
        _m_status_cache_file,
        _m_should_reboot,
        _m_remove_notice,
        m_on_supported_kernel,
        avail_res,
        entitled_res,
        uf_entitled,
        uf_status,
        FakeConfig,
        fake_machine_token_file,
    ):
        """Test we get the correct status dict when attached with basic conf"""
        resource_names = [resource["name"] for resource in avail_res]
        default_entitled = ContractStatus.UNENTITLED.value
        default_status = UserFacingStatus.UNAVAILABLE.value
        token = {
            "availableResources": [],
            "machineTokenInfo": {
                "machineId": "test_machine_id",
                "accountInfo": {
                    "id": "acct-1",
                    "name": "test_account",
                    "createdAt": datetime.datetime(
                        2019,
                        6,
                        14,
                        6,
                        45,
                        50,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "externalAccountIDs": [{"IDs": ["id1"], "origin": "AWS"}],
                },
                "contractInfo": {
                    "id": "cid",
                    "name": "test_contract",
                    "createdAt": datetime.datetime(
                        2020,
                        5,
                        8,
                        19,
                        2,
                        26,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "effectiveFrom": datetime.datetime(
                        2000,
                        5,
                        8,
                        19,
                        2,
                        26,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "effectiveTo": datetime.datetime(
                        2040,
                        5,
                        8,
                        19,
                        2,
                        26,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "resourceEntitlements": entitled_res,
                    "products": ["free"],
                },
            },
        }

        available_resource_response = [
            {
                "name": cls.name,
                "available": bool(
                    {"name": cls.name, "available": True} in avail_res
                ),
            }
            for cls in ENTITLEMENT_CLASSES
        ]
        if avail_res:
            token["availableResources"] = available_resource_response
        else:
            m_get_avail_resources.return_value = available_resource_response

        cfg = FakeConfig()

        fake_machine_token_file.attached = True
        fake_machine_token_file.token = token

        expected_services = [
            {
                "description": cls.description,
                "entitled": (
                    uf_entitled
                    if cls.name in resource_names
                    else default_entitled
                ),
                "name": cls.name,
                "status": (
                    uf_status if cls.name in resource_names else default_status
                ),
                "status_details": mock.ANY,
                "description_override": None,
                "available": mock.ANY,
                "blocked_by": [],
                "warning": None,
                "variants": {},
            }
            for cls in ENTITLEMENT_CLASSES
        ]
        expected_services.sort(key=lambda x: x.get("name", ""))
        expected = copy.deepcopy(DEFAULT_STATUS)
        expected.update(
            {
                "version": mock.ANY,
                "attached": True,
                "machine_id": "test_machine_id",
                "services": expected_services,
                "effective": datetime.datetime(
                    2000, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
                ),
                "expires": datetime.datetime(
                    2040, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
                ),
                "contract": {
                    "name": "test_contract",
                    "id": "cid",
                    "created_at": datetime.datetime(
                        2020, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
                    ),
                    "products": ["free"],
                    "tech_support_level": "n/a",
                },
                "account": {
                    "name": "test_account",
                    "id": "acct-1",
                    "created_at": datetime.datetime(
                        2019, 6, 14, 6, 45, 50, tzinfo=datetime.timezone.utc
                    ),
                    "external_account_ids": [
                        {"IDs": ["id1"], "origin": "AWS"}
                    ],
                },
            }
        )
        with mock.patch(
            "uaclient.status._get_config_status"
        ) as m_get_cfg_status:
            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
            assert expected == status.status(cfg=cfg, show_all=True)
        if avail_res:
            assert m_get_avail_resources.call_count == 0
        else:
            assert m_get_avail_resources.call_count == 1
        # status() idempotent
        with mock.patch(
            "uaclient.status._get_config_status"
        ) as m_get_cfg_status:
            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
            assert expected == status.status(cfg=cfg, show_all=True)

    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.util.we_are_currently_root")
    @mock.patch("uaclient.status.get_available_resources")
    @mock.patch(
        "uaclient.files.user_config_file.UserConfigFileObject.public_config",
        new_callable=mock.PropertyMock,
    )
    def test_nonroot_unattached_is_same_as_unattached_root(
        self,
        m_public_config,
        m_get_available_resources,
        m_we_are_currently_root,
        _m_status_cache_file,
        _m_should_reboot,
        _m_remove_notice,
        m_on_supported_kernel,
        FakeConfig,
    ):
        m_public_config.return_value = UserConfigData()
        m_get_available_resources.return_value = [
            {"name": "esm-infra", "available": True}
        ]
        m_we_are_currently_root.return_value = False
        cfg = FakeConfig()
        nonroot_status = status.status(cfg=cfg)

        m_we_are_currently_root.return_value = True
        cfg = FakeConfig()
        root_unattached_status = status.status(cfg=cfg)

        assert root_unattached_status == nonroot_status

    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.util.we_are_currently_root")
    @mock.patch("uaclient.status.get_available_resources")
    @mock.patch(
        "uaclient.files.user_config_file.UserConfigFileObject.public_config",
        new_callable=mock.PropertyMock,
    )
    def test_root_and_non_root_are_same_attached(
        self,
        m_public_config,
        m_get_available_resources,
        m_we_are_currently_root,
        _m_status_cache_file,
        _m_should_reboot,
        _m_remove_notice,
        m_on_supported_kernel,
        FakeConfig,
        fake_machine_token_file,
    ):
        m_public_config.return_value = UserConfigData()
        m_we_are_currently_root.return_value = True
        fake_machine_token_file.attached = True
        cfg = FakeConfig()
        root_status = status.status(cfg=cfg)
        m_we_are_currently_root.return_value = False
        normal_status = status.status(cfg=cfg)
        assert normal_status == root_status

    @pytest.mark.parametrize("variants_in_contract", ((True), (False)))
    @pytest.mark.parametrize("show_all", (True, False))
    @pytest.mark.parametrize(
        "entitlements",
        (
            [],
            [
                {
                    "type": "support",
                    "entitled": True,
                    "affordances": {"supportLevel": "anything"},
                }
            ],
        ),
    )
    @pytest.mark.usefixtures("all_resources_available")
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch(
        M_PATH + "fips.FIPSCommonEntitlement.application_status",
        return_value=(ApplicationStatus.DISABLED, ""),
    )
    @mock.patch(
        M_PATH + "livepatch.LivepatchEntitlement.application_status",
        return_value=(ApplicationStatus.DISABLED, ""),
    )
    @mock.patch(M_PATH + "livepatch.LivepatchEntitlement.user_facing_status")
    @mock.patch(M_PATH + "livepatch.LivepatchEntitlement.contract_status")
    @mock.patch(M_PATH + "landscape.LandscapeEntitlement.user_facing_status")
    @mock.patch(M_PATH + "landscape.LandscapeEntitlement.contract_status")
    @mock.patch(M_PATH + "esm.ESMAppsEntitlement.user_facing_status")
    @mock.patch(M_PATH + "esm.ESMAppsEntitlement.contract_status")
    @mock.patch(M_PATH + "repo.RepoEntitlement.user_facing_status")
    @mock.patch(M_PATH + "repo.RepoEntitlement.contract_status")
    @mock.patch(M_PATH + "base.UAEntitlement._get_contract_variants")
    def test_attached_reports_contract_and_service_status(
        self,
        m_contract_variants,
        m_repo_contract_status,
        m_repo_uf_status,
        m_esm_contract_status,
        m_esm_uf_status,
        m_landscape_contract_status,
        m_landscape_uf_status,
        m_livepatch_contract_status,
        m_livepatch_uf_status,
        _m_livepatch_status,
        _m_fips_status,
        _m_status_cache_file,
        _m_should_reboot,
        _m_remove_notice,
        m_on_supported_kernel,
        all_resources_available,
        entitlements,
        show_all,
        variants_in_contract,
        FakeConfig,
        fake_machine_token_file,
    ):
        """When attached, return contract and service user-facing status."""
        m_repo_contract_status.return_value = ContractStatus.ENTITLED
        m_repo_uf_status.return_value = (
            UserFacingStatus.INAPPLICABLE,
            messages.NamedMessage("test-code", "repo details"),
        )
        m_livepatch_contract_status.return_value = ContractStatus.ENTITLED
        m_livepatch_uf_status.return_value = (
            UserFacingStatus.ACTIVE,
            messages.NamedMessage("test-code", "livepatch details"),
        )
        m_esm_contract_status.return_value = ContractStatus.ENTITLED
        m_esm_uf_status.return_value = (
            UserFacingStatus.INAPPLICABLE,
            messages.NamedMessage("test-code", "esm-apps details"),
        )
        m_landscape_contract_status.return_value = ContractStatus.ENTITLED
        m_landscape_uf_status.return_value = (
            UserFacingStatus.INAPPLICABLE,
            messages.NamedMessage("test-code", "landscape details"),
        )
        token = {
            "availableResources": all_resources_available,
            "machineTokenInfo": {
                "machineId": "test_machine_id",
                "accountInfo": {
                    "id": "1",
                    "name": "accountname",
                    "createdAt": datetime.datetime(
                        2019, 6, 14, 6, 45, 50, tzinfo=datetime.timezone.utc
                    ),
                    "externalAccountIDs": [{"IDs": ["id1"], "origin": "AWS"}],
                },
                "contractInfo": {
                    "id": "contract-1",
                    "name": "contractname",
                    "createdAt": datetime.datetime(
                        2020, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
                    ),
                    "resourceEntitlements": entitlements,
                    "products": ["free"],
                },
            },
        }

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        fake_machine_token_file.token = token

        mock_notice = NoticesManager()
        if not entitlements:
            support_level = UserFacingStatus.INAPPLICABLE.value
        else:
            support_level = entitlements[0]["affordances"]["supportLevel"]
        expected = copy.deepcopy(status.DEFAULT_STATUS)
        expected.update(
            {
                "version": mock.ANY,
                "attached": True,
                "machine_id": "test_machine_id",
                "contract": {
                    "name": "contractname",
                    "id": "contract-1",
                    "created_at": datetime.datetime(
                        2020, 5, 8, 19, 2, 26, tzinfo=datetime.timezone.utc
                    ),
                    "products": ["free"],
                    "tech_support_level": support_level,
                },
                "account": {
                    "name": "accountname",
                    "id": "1",
                    "created_at": datetime.datetime(
                        2019, 6, 14, 6, 45, 50, tzinfo=datetime.timezone.utc
                    ),
                    "external_account_ids": [
                        {"IDs": ["id1"], "origin": "AWS"}
                    ],
                },
            }
        )

        ent_details = {
            "livepatch": "livepatch details",
            "esm-apps": "esm-apps details",
            "landscape": "landscape details",
        }

        rt_variants = {
            "intel-iotg": {
                "available": mock.ANY,
                "blocked_by": [],
                "description": "RT kernel optimized "
                "for Intel IOTG "
                "platform",
                "description_override": None,
                "entitled": "yes",
                "name": "intel-iotg",
                "status": "n/a",
                "status_details": "repo details",
                "warning": None,
            },
            "nvidia-tegra": {
                "available": mock.ANY,
                "blocked_by": [],
                "description": "RT kernel "
                "optimized for "
                "NVIDIA Tegra "
                "platform",
                "description_override": None,
                "entitled": "yes",
                "name": "nvidia-tegra",
                "status": "n/a",
                "status_details": "repo details",
                "warning": None,
            },
            "generic": {
                "available": mock.ANY,
                "blocked_by": [],
                "description": "Generic version of the RT kernel (default)",
                "description_override": None,
                "entitled": "yes",
                "name": "generic",
                "status": "n/a",
                "status_details": "repo details",
                "warning": None,
            },
        }

        for cls in ENTITLEMENT_CLASSES:
            if cls.name == "realtime-kernel":
                if variants_in_contract:
                    m_contract_variants.return_value = set(
                        ["intel-iotg", "nvidia-tegra"]
                    )
                    variants = rt_variants
                else:
                    m_contract_variants.return_value = set()
                    variants = {}
            else:
                variants = {}

            if cls.name == "livepatch":
                expected_status = UserFacingStatus.ACTIVE.value
            elif show_all:
                expected_status = UserFacingStatus.INAPPLICABLE.value
            else:
                continue

            expected["services"].append(
                {
                    "name": cls.name,
                    "description": cls.description,
                    "entitled": ContractStatus.ENTITLED.value,
                    "status": expected_status,
                    "status_details": ent_details.get(
                        cls.name, "repo details"
                    ),
                    "description_override": None,
                    "available": mock.ANY,
                    "blocked_by": [],
                    "warning": None,
                    "variants": variants,
                }
            )
        expected["services"].sort(key=lambda x: x.get("name", ""))
        with mock.patch(
            "uaclient.status._get_config_status"
        ) as m_get_cfg_status:
            m_get_cfg_status.return_value = DEFAULT_CFG_STATUS
            expected_status_calls = 15 if variants_in_contract else 12
            assert expected == status.status(cfg=cfg, show_all=show_all)
            assert expected_status_calls == m_repo_uf_status.call_count
            assert 1 == m_livepatch_uf_status.call_count

        expected_calls = [
            mock.call(Notice.AUTO_ATTACH_RETRY_FULL_NOTICE),
            mock.call(Notice.AUTO_ATTACH_RETRY_TOTAL_FAILURE),
        ]

        assert expected_calls == mock_notice.remove.call_args_list

    @pytest.mark.usefixtures("all_resources_available")
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.util.we_are_currently_root")
    @mock.patch("uaclient.status.get_available_resources")
    @mock.patch(
        "uaclient.files.user_config_file.UserConfigFileObject.public_config",
        new_callable=mock.PropertyMock,
    )
    def test_expires_handled_appropriately(
        self,
        m_public_config,
        _m_get_available_resources,
        m_we_are_currently_root,
        _m_status_cache_file,
        _m_should_reboot,
        _m_remove_notice,
        m_on_supported_kernel,
        all_resources_available,
        FakeConfig,
        fake_machine_token_file,
    ):
        m_public_config.return_value = UserConfigData()
        m_we_are_currently_root.return_value = True
        token = {
            "availableResources": all_resources_available,
            "machineTokenInfo": {
                "machineId": "test_machine_id",
                "accountInfo": {"id": "1", "name": "accountname"},
                "contractInfo": {
                    "name": "contractname",
                    "id": "contract-1",
                    "effectiveTo": datetime.datetime(
                        2020, 7, 18, 0, 0, 0, tzinfo=datetime.timezone.utc
                    ),
                    "createdAt": datetime.datetime(
                        2020, 5, 8, 0, 0, 0, tzinfo=datetime.timezone.utc
                    ),
                    "resourceEntitlements": [],
                    "products": ["free"],
                },
            },
        }
        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        fake_machine_token_file.token = token

        # Test that root's status works as expected (including the cache write)
        expected_dt = datetime.datetime(
            2020, 7, 18, 0, 0, 0, tzinfo=datetime.timezone.utc
        )
        assert expected_dt == status.status(cfg=cfg)["expires"]

        # Test that the read from the status cache work properly for non-root
        # users
        m_we_are_currently_root.return_value = False
        cfg = FakeConfig()
        assert expected_dt == status.status(cfg=cfg)["expires"]

    @mock.patch(
        "uaclient.files.state_files.reboot_cmd_marker_file",
        new_callable=mock.PropertyMock,
    )
    @mock.patch(
        "uaclient.files.user_config_file.UserConfigFileObject.public_config",
        new_callable=mock.PropertyMock,
    )
    @mock.patch("uaclient.files.state_files.status_cache_file.read")
    @mock.patch("uaclient.files.state_files.status_cache_file.write")
    @mock.patch("uaclient.status.get_available_resources", return_value={})
    def test_nonroot_user_does_not_use_cache(
        self,
        _m_get_available_resources,
        _m_status_cache_file_write,
        m_status_cache_file_read,
        m_public_config,
        m_reboot_cmd_marker_file,
        _m_should_reboot,
        m_remove_notice,
        m_on_supported_kernel,
        FakeConfig,
    ):
        m_public_config.return_value = UserConfigData()
        m_reboot_cmd_marker_file.is_present = True
        cached_status = {"pass": True}
        m_status_cache_file_read.return_value = cached_status
        cfg = FakeConfig()
        before = status.status(cfg=cfg)

        # Even non-root users can update execution_status details
        details = messages.ENABLE_REBOOT_REQUIRED_TMPL.format(
            operation="configuration changes"
        )
        reboot_required = UserFacingConfigStatus.REBOOTREQUIRED.value
        cached_status.update(
            {
                "execution_status": reboot_required,
                "execution_details": details,
                "features": {},
                "notices": [],
                "config_path": None,
                "config": {"data_dir": mock.ANY},
                "services": [],
            }
        )

        assert cached_status != status.status(cfg=cfg)
        assert before == status.status(cfg=cfg)


ATTACHED_SERVICE_STATUS_PARAMETERS = [
    # ENTITLED => display the given user-facing status
    (ContractStatus.ENTITLED, UserFacingStatus.ACTIVE, False, "enabled"),
    (ContractStatus.ENTITLED, UserFacingStatus.INACTIVE, False, "disabled"),
    (ContractStatus.ENTITLED, UserFacingStatus.INAPPLICABLE, False, "n/a"),
    (ContractStatus.ENTITLED, UserFacingStatus.UNAVAILABLE, False, "—"),
    # UNENTITLED => UNAVAILABLE
    (ContractStatus.UNENTITLED, UserFacingStatus.ACTIVE, False, "—"),
    (ContractStatus.UNENTITLED, UserFacingStatus.INACTIVE, False, "—"),
    (ContractStatus.UNENTITLED, UserFacingStatus.INAPPLICABLE, False, "—"),
    (ContractStatus.UNENTITLED, UserFacingStatus.UNAVAILABLE, [], "—"),
    # ENTITLED but in unavailable_resources => INAPPLICABLE
    (ContractStatus.ENTITLED, UserFacingStatus.ACTIVE, True, "n/a"),
    (ContractStatus.ENTITLED, UserFacingStatus.INACTIVE, True, "n/a"),
    (ContractStatus.ENTITLED, UserFacingStatus.INAPPLICABLE, True, "n/a"),
    (ContractStatus.ENTITLED, UserFacingStatus.UNAVAILABLE, True, "n/a"),
    # UNENTITLED and in unavailable_resources => UNAVAILABLE
    (ContractStatus.UNENTITLED, UserFacingStatus.ACTIVE, True, "—"),
    (ContractStatus.UNENTITLED, UserFacingStatus.INACTIVE, True, "—"),
    (ContractStatus.UNENTITLED, UserFacingStatus.INAPPLICABLE, True, "—"),
    (ContractStatus.UNENTITLED, UserFacingStatus.UNAVAILABLE, True, "—"),
]


class TestAttachedServiceStatus:
    @pytest.mark.parametrize(
        "contract_status,uf_status,in_inapplicable_resources,expected_status",
        ATTACHED_SERVICE_STATUS_PARAMETERS,
    )
    def test_status(
        self,
        contract_status,
        uf_status,
        in_inapplicable_resources,
        expected_status,
        FakeConfig,
    ):
        ent = mock.MagicMock()
        ent.name = "test_entitlement"
        ent.contract_status.return_value = contract_status
        ent.user_facing_status.return_value = (
            uf_status,
            messages.NamedMessage("test-code", ""),
        )

        unavailable_resources = (
            {ent.name: ""} if in_inapplicable_resources else {}
        )
        ret = status._attached_service_status(
            ent, unavailable_resources, FakeConfig()
        )

        assert expected_status == ret["status"]

    @pytest.mark.parametrize(
        "blocking_incompatible_services, expected_blocked_by",
        (
            ([], []),
            (
                [
                    EntitlementWithMessage(
                        FIPSEntitlement, messages.NamedMessage("code", "msg")
                    )
                ],
                [{"name": "fips", "reason": "msg", "reason_code": "code"}],
            ),
            (
                [
                    EntitlementWithMessage(
                        FIPSEntitlement, messages.NamedMessage("code", "msg")
                    ),
                    EntitlementWithMessage(
                        ROSEntitlement, messages.NamedMessage("code2", "msg2")
                    ),
                ],
                [
                    {"name": "fips", "reason": "msg", "reason_code": "code"},
                    {"name": "ros", "reason": "msg2", "reason_code": "code2"},
                ],
            ),
        ),
    )
    def test_blocked_by(
        self,
        blocking_incompatible_services,
        expected_blocked_by,
        tmpdir,
        FakeConfig,
    ):
        cfg = FakeConfig()
        ent = ConcreteTestEntitlement(
            cfg=cfg,
            blocking_incompatible_services=blocking_incompatible_services,
        )
        service_status = status._attached_service_status(ent, [], cfg)
        assert service_status["blocked_by"] == expected_blocked_by



import logging
import os
from typing import Optional
from urllib.parse import urlparse

from uaclient import defaults, event_logger, util
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    EnumDataValue,
    Field,
    IntDataValue,
    StringDataValue,
)
from uaclient.files.data_types import DataObjectFile, DataObjectFileFormat
from uaclient.files.files import UAFile

# Config proxy fields that are visible and configurable
PROXY_FIELDS = [
    "apt_http_proxy",
    "apt_https_proxy",
    "global_apt_http_proxy",
    "global_apt_https_proxy",
    "ua_apt_http_proxy",
    "ua_apt_https_proxy",
    "http_proxy",
    "https_proxy",
]


class LXDGuestAttachEnum(EnumDataValue):
    ON = "on"
    OFF = "off"
    AVAILABLE = "available"

    def __str__(self):
        return self.value


class UserConfigData(DataObject):
    fields = [
        Field("apt_http_proxy", StringDataValue, required=False),
        Field("apt_https_proxy", StringDataValue, required=False),
        Field("global_apt_http_proxy", StringDataValue, required=False),
        Field("global_apt_https_proxy", StringDataValue, required=False),
        Field("ua_apt_http_proxy", StringDataValue, required=False),
        Field("ua_apt_https_proxy", StringDataValue, required=False),
        Field("http_proxy", StringDataValue, required=False),
        Field("https_proxy", StringDataValue, required=False),
        Field("apt_news", BoolDataValue, required=False),
        Field("apt_news_url", StringDataValue, required=False),
        Field("poll_for_pro_license", BoolDataValue, required=False),
        Field("polling_error_retry_delay", IntDataValue, required=False),
        Field("metering_timer", IntDataValue, required=False),
        Field("update_messaging_timer", IntDataValue, required=False),
        Field(
            "vulnerability_data_url_prefix", StringDataValue, required=False
        ),
        Field("lxd_guest_attach", LXDGuestAttachEnum, required=False),
    ]

    def __init__(
        self,
        apt_http_proxy: Optional[str] = None,
        apt_https_proxy: Optional[str] = None,
        global_apt_http_proxy: Optional[str] = None,
        global_apt_https_proxy: Optional[str] = None,
        ua_apt_http_proxy: Optional[str] = None,
        ua_apt_https_proxy: Optional[str] = None,
        http_proxy: Optional[str] = None,
        https_proxy: Optional[str] = None,
        apt_news: Optional[bool] = None,
        apt_news_url: Optional[str] = None,
        poll_for_pro_license: Optional[bool] = None,
        polling_error_retry_delay: Optional[int] = None,
        metering_timer: Optional[int] = None,
        update_messaging_timer: Optional[int] = None,
        vulnerability_data_url_prefix: Optional[str] = None,
        lxd_guest_attach: Optional[LXDGuestAttachEnum] = None,
    ):
        self.apt_http_proxy = apt_http_proxy
        self.apt_https_proxy = apt_https_proxy
        self.global_apt_http_proxy = global_apt_http_proxy
        self.global_apt_https_proxy = global_apt_https_proxy
        self.ua_apt_http_proxy = ua_apt_http_proxy
        self.ua_apt_https_proxy = ua_apt_https_proxy
        self.http_proxy = http_proxy
        self.https_proxy = https_proxy
        self.apt_news = apt_news
        self.apt_news_url = apt_news_url
        self.poll_for_pro_license = poll_for_pro_license
        self.polling_error_retry_delay = polling_error_retry_delay
        self.metering_timer = metering_timer
        self.update_messaging_timer = update_messaging_timer
        self.vulnerability_data_url_prefix = vulnerability_data_url_prefix
        self.lxd_guest_attach = lxd_guest_attach


event = event_logger.get_event_logger()
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class UserConfigFileObject:
    def __init__(self, directory: str = defaults.DEFAULT_DATA_DIR):
        file_name = defaults.USER_CONFIG_FILE
        self._private = DataObjectFile(
            UserConfigData,
            UAFile(
                file_name,
                os.path.join(directory, defaults.PRIVATE_SUBDIR),
                private=True,
            ),
            DataObjectFileFormat.JSON,
            optional_type_errors_become_null=True,
        )
        self._public = DataObjectFile(
            UserConfigData,
            UAFile(file_name, directory, private=False),
            DataObjectFileFormat.JSON,
            optional_type_errors_become_null=True,
        )

    @property
    def public_config(self) -> UserConfigData:
        public_config = self._public.read()
        if public_config is None:
            public_config = UserConfigData()
        return public_config

    def redact_config_data(
        self, user_config: UserConfigData
    ) -> UserConfigData:
        redacted_data_dict = user_config.to_dict()
        for field in PROXY_FIELDS:
            value = redacted_data_dict.get(field)
            if value:
                parsed_url = urlparse(value)
                if parsed_url.username or parsed_url.password:
                    redacted_data_dict[field] = "<REDACTED>"
        return UserConfigData.from_dict(redacted_data_dict)

    def read(self) -> UserConfigData:
        if util.we_are_currently_root():
            private_config = self._private.read()
            if private_config is not None:
                return private_config
        public_config = self._public.read()
        if public_config is not None:
            return public_config
        return UserConfigData()

    def write(self, content: UserConfigData):
        self._private.write(content)
        redacted_content = self.redact_config_data(content)
        self._public.write(redacted_content)


user_config = UserConfigFileObject(defaults.DEFAULT_DATA_DIR)



import datetime
import io
import logging
import os
import shutil
from enum import Enum

import mock
import pytest

# Useless try/except to make flake8 happy \_("/)_/
try:
    from uaclient import event_logger
    from uaclient.config import UAConfig
    from uaclient.entitlements.entitlement_status import ApplicationStatus
    from uaclient.files.machine_token import MachineTokenFile
    from uaclient.files.notices import NoticeFileDetails
    from uaclient.files.user_config_file import UserConfigData
except ImportError:
    raise


shutil.get_terminal_size = mock.MagicMock(
    return_value=os.terminal_size((80, 20))
)


@pytest.yield_fixture(scope="session", autouse=True)
def _subp():
    """
    A fixture that mocks system._subp for all tests.
    If a test needs the actual _subp, this fixture yields it,
    so just add an argument to the test named "_subp".
    """
    from uaclient.system import _subp

    original = _subp
    with mock.patch(
        "uaclient.system._subp", return_value=("mockstdout", "mockstderr")
    ):
        yield original


@pytest.yield_fixture(scope="session", autouse=True)
def _warn_about_new_version():
    """
    A fixture that mocks cli._warn_about_new_version for all tests.
    If a test needs the actual _warn_about_new_version, this fixture yields it,
    so just add an argument to the test named "_warn_about_new_version".
    """
    from uaclient.cli import _warn_about_new_version

    original = _warn_about_new_version
    with mock.patch("uaclient.cli._warn_about_new_version"):
        yield original


@pytest.yield_fixture(scope="session", autouse=True)
def update_sources_list():
    """
    A fixture that mocks apt.update_sources_list for all tests.
    If a test needs the actual update_sources_list, this fixture yields it,
    so just add an argument to the test named "update_sources_list".
    """
    from uaclient.apt import update_sources_list

    original = update_sources_list
    with mock.patch("uaclient.apt.update_sources_list"):
        yield original


@pytest.yield_fixture(scope="session", autouse=True)
def util_we_are_currently_root():
    """
    A fixture that mocks util.we_are_currently_root for all tests.
    Default to true as most tests need it to be true.
    """
    from uaclient.util import we_are_currently_root

    original = we_are_currently_root
    with mock.patch("uaclient.util.we_are_currently_root", return_value=True):
        yield original


@pytest.yield_fixture(scope="session", autouse=True)
def urllib_request_urlopen():
    """
    A fixture that mocks urlopen for all tests.
    This prevents us from accidentally making requests in unit tests
    """
    from urllib.request import urlopen

    original = urlopen
    with mock.patch("urllib.request.urlopen"):
        yield original


@pytest.fixture
def caplog_text(request):
    """
    A fixture that returns a function that returns caplog.text

    caplog isn't available in pytest in all of our target releases; this either
    uses caplog.text if available, or a shim which replicates what it does.

    Specifically, bionic is the first Ubuntu release to contain a version of
    pytest new enough for the caplog fixture to be present. In xenial, the
    python3-pytest-catchlog package provides the same functionality (this is
    the code that was later integrated in to pytest).

    (It returns a function so that the requester can decide when to examine the
    logs; if it returned caplog.text directly, that would always be empty.)
    """
    cap_params = getattr(request, "param", logging.INFO)
    log_filter = None
    if isinstance(cap_params, tuple):
        log_level = cap_params[0]
        log_filter = cap_params[1]
    else:
        log_level = cap_params
    try:
        try:
            caplog = request.getfixturevalue("caplog")
        except AttributeError:
            # Older versions of pytest only have getfuncargvalue, which is now
            # deprecated in favour of getfixturevalue
            caplog = request.getfuncargvalue("caplog")
        caplog.set_level(log_level)
        if log_filter:
            caplog.handler.addFilter(log_filter())

        def _func():
            return caplog.text

    except LookupError:
        # If the caplog fixture isn't available, shim something in ourselves
        root = logging.getLogger("ubuntupro")
        root.propagate = False
        root.setLevel(log_level)
        handler = logging.StreamHandler(io.StringIO())
        handler.setFormatter(
            logging.Formatter(
                "%(filename)-25s %(lineno)4d %(levelname)-8s %(message)s"
            )
        )
        if log_filter:
            handler.addFilter(log_filter())
        root.addHandler(handler)

        def _func():
            return handler.stream.getvalue()

        def clear_handlers():
            root.handlers = []

        request.addfinalizer(clear_handlers)
    return _func


@pytest.yield_fixture
def logging_sandbox():
    # Monkeypatch a replacement root logger, so that our changes to logging
    # configuration don't persist outside of the test
    upro_logger = logging.getLogger("ubuntupro")
    upro_logger.setLevel(logging.WARNING)

    with mock.patch.object(logging, "root", upro_logger):
        with mock.patch.object(logging.Logger, "root", upro_logger):
            with mock.patch.object(
                logging.Logger, "manager", logging.Manager(upro_logger)
            ):
                yield


@pytest.yield_fixture(scope="function", autouse=True)
def fake_machine_token_file():
    from unittest.mock import patch

    with patch(
        "uaclient.files.machine_token.get_machine_token_file"
    ) as m_get_machine_token_file:
        machine_token = FakeMachineToken(attached=False)
        m_get_machine_token_file.return_value = machine_token
        yield machine_token


class FakeMachineToken(MachineTokenFile):
    def __init__(self, attached, token=None):
        self.attached = attached
        self._machine_token = None
        self.token = token
        self.machine_token_overlay_path = None
        self._contract_expiry_datetime = None
        self._entitlements = None
        self.write_calls = 0
        self.delete_calls = 0

    @property
    def machine_token(self):
        self._machine_token = None
        return super().machine_token

    @property
    def is_present(self):
        return self.attached

    def read(self):
        if self.token:
            return self.token

        if not self.attached:
            return None

        return {
            "availableResources": [],
            "machineToken": "not-null",
            "machineTokenInfo": {
                "machineId": "test_machine_id",
                "accountInfo": {
                    "id": "acct-1",
                    "name": "test",
                    "createdAt": datetime.datetime(
                        2019,
                        6,
                        14,
                        6,
                        45,
                        50,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "externalAccountIDs": [{"IDs": ["id1"], "origin": "AWS"}],
                },
                "contractInfo": {
                    "id": "cid",
                    "name": "test_contract",
                    "createdAt": datetime.datetime(
                        2020,
                        5,
                        8,
                        19,
                        2,
                        26,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "effectiveFrom": datetime.datetime(
                        2000,
                        5,
                        8,
                        19,
                        2,
                        26,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "effectiveTo": datetime.datetime(
                        2040,
                        5,
                        8,
                        19,
                        2,
                        26,
                        tzinfo=datetime.timezone.utc,
                    ),
                    "resourceEntitlements": [],
                    "products": ["free"],
                },
            },
        }

    def write(self, private_content):
        self.token = private_content
        self.write_calls += 1

    def delete(self):
        self.delete_calls += 1


@pytest.fixture
def FakeConfig(tmpdir):
    class _FakeConfig(UAConfig):
        def __init__(
            self,
            cfg_overrides={},
            features_override=None,
        ) -> None:
            if not cfg_overrides.get("data_dir"):
                cfg_overrides.update({"data_dir": tmpdir.strpath})
            if not cfg_overrides.get("log_file"):
                cfg_overrides.update(
                    {"log_file": tmpdir.join("log_file.log").strpath}
                )
            super().__init__(
                cfg_overrides,
                user_config=UserConfigData(),
            )

        def override_features(self, features_override):
            if features_override is not None:
                self.cfg.update({"features": features_override})

    return _FakeConfig


@pytest.fixture
def event():
    event = event_logger.get_event_logger()
    event.reset()

    return event


class FakeNotice(NoticeFileDetails, Enum):
    reboot_required = NoticeFileDetails(
        "10", "reboot_required", False, "notice_a"
    )
    reboot_script_failed = NoticeFileDetails(
        "12", "reboot_script_failed", True, "notice_a2"
    )
    enable_reboot_required = NoticeFileDetails(
        "11", "enable_reboot_required", False, "notice_b"
    )


@pytest.yield_fixture(autouse=True)
def mock_notices_dir(tmpdir_factory):
    perm_dir = tmpdir_factory.mktemp("notices")
    temp_dir = tmpdir_factory.mktemp("temp_notices")
    with mock.patch(
        "uaclient.defaults.NOTICES_PERMANENT_DIRECTORY",
        perm_dir.strpath,
    ):
        with mock.patch(
            "uaclient.defaults.NOTICES_TEMPORARY_DIRECTORY",
            temp_dir.strpath,
        ):
            yield


@pytest.fixture
def mock_entitlement():
    def factory_func(
        *,
        name="test",
        title="test",
        application_status=(ApplicationStatus.DISABLED, ""),
        can_disable=(False, None),
        disable=(False, None),
        enable=(False, None)
    ):
        m_ent_cls = mock.MagicMock()
        type(m_ent_cls).name = mock.PropertyMock(return_value=name)
        m_ent_obj = m_ent_cls.return_value
        type(m_ent_obj).title = mock.PropertyMock(return_value=title)
        m_ent_obj.application_status.return_value = application_status
        m_ent_obj.disable.return_value = disable
        m_ent_obj.enable.return_value = enable
        m_ent_obj.can_disable.return_value = can_disable

        return (m_ent_cls, m_ent_obj)

    return factory_func
